"""Error for artefact results."""
from __future__ import annotations

# std
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Optional, Type, TypeVar, Union

# external
from redis import Redis

# module
from ._constants import ARTEFACTS, hash_t, join


class UnwrapError(Exception):
    """Exception thrown when unwrapping an error."""

    pass


class ErrorKind(str, Enum):
    """Kinds of errors."""

    # db errors
    NotFound = "NotFound"
    Mismatch = "Mismatch"
    UnresolvedLink = "UnresolvedLink"
    # Storage error
    DataNotFound = "DataNotFound"
    # Type errors
    WrongType = "WrongType"
    JSONEncodingError = "JSONEncodingError"
    JSONDecodingError = "JSONDecodingError"
    UnknownEncodingError = "UnknownEncodingError"
    # Job error conditions
    MissingOutput = "MissingOutput"
    MissingInput = "MissingInput"
    ExceptionRaised = "ExceptionRaised"
    JobTimedOut = "JobTimedOut"
    NoErrorData = "NoErrorData"
    KilledBySignal = "KilledBySignal"


@dataclass
class Error:
    """An Error value for artefacts."""

    kind: ErrorKind
    source: Optional[hash_t] = None
    details: Optional[str] = None

    def put(self: "Error", db: Redis[bytes], hash: hash_t) -> None:
        """Save an Error to Redis."""
        data = dict(kind=self.kind.name)
        if self.source:
            data["source"] = str(self.source)
        if self.details:
            data["details"] = str(self.details)
        db.hset(
            join(ARTEFACTS, hash, "error"),
            mapping=data,  # type:ignore
        )

    @classmethod
    def grab(cls: Type["Error"], db: Redis[bytes], hash: hash_t) -> "Error":
        """Grab an Error from the Redis store."""
        if not join(ARTEFACTS, hash, "error"):
            raise RuntimeError(f"No error for artefact at {hash}")

        data = db.hgetall(join(ARTEFACTS, hash, "error"))
        kind = ErrorKind(data[b"kind"].decode())

        # Sometimes the python boilerplate is really freaking annoying...
        tmp = data.get(b"source", None)
        if tmp is not None:
            source: Optional[hash_t] = hash_t(tmp.decode())
        else:
            source = None
        tmp = data.get(b"details", None)
        if tmp is not None:
            details: Optional[str] = tmp.decode()
        else:
            details = None

        return Error(kind, source=source, details=details)


# A simple mypy result type
T = TypeVar("T")
Result = Union[Error, T]
"""Result contains either a value or an `Error` instance.

`Result[T]` is a type hint that corresponds to `Union[T, Error]`. This is only
a an abstraction: at runtime, a `Result[T]` is just `T` or `Error`. That is,
`Result` has no runtime representation.

"""


def unwrap(it: Result[T]) -> T:
    """Unwrap a `errors.Result` type.

    Unwrap `errors.Result[T]` and return `T`. If `errors.Result[T]` is of type
    `Error`, this function raises `errors.UnwrapError`.

    Args:
        it: An object of type `errors.Result[T]`.

    Returns:
        The value of it with type `T`.

    Raises:
        UnwrapError: `errors.Result[T]` is an `errors.Error` instance.

    """
    if isinstance(it, Error):
        raise UnwrapError(
            f"data is errored: kind={it.kind}"
            + f"\nsource={it.source}"
            + f"\ndetails={it.details}"
        )
    else:
        return it


T1 = TypeVar("T1")
T2 = TypeVar("T2")


def match(
    result: Result[T], some: Callable[[T], T1], none: Callable[[Error], T2]
) -> Union[T1, T2]:
    """Pattern matching on Results."""
    if isinstance(result, Error):
        return none(result)
    else:
        return some(result)



"""Functions for describing redis-backed DAGs."""
from __future__ import annotations

# std
from contextlib import ContextDecorator
from enum import IntEnum
from io import BytesIO
import signal
import traceback
from types import FrameType
from typing import Any, Dict, Optional, Union

# external
from redis import Redis
import rq

# module
from . import _serdes
from ._constants import ARTEFACTS, hash_t, join, OPERATIONS
from ._context import _options_stack
from ._funsies import Funsie, FunsieHow
from ._graph import (
    Artefact,
    ArtefactStatus,
    create_link,
    get_status,
    get_stream,
    mark_error,
    Operation,
    resolve_link,
    set_data,
)
from ._logging import logger
from ._pyfunc import run_python_funsie  # runner for python functions
from ._shell import run_shell_funsie  # runner for shell
from ._storage import StorageEngine
from ._subdag import run_subdag_funsie  # runner for shell
from .errors import Error, ErrorKind, Result

# Dictionary of runners
RUNNERS = {
    FunsieHow.shell: run_shell_funsie,
    FunsieHow.python: run_python_funsie,
    FunsieHow.subdag: run_subdag_funsie,
}
out_data_t = Union[Dict[str, Optional[object]], Dict[str, Optional[Artefact[Any]]]]


# ----------------------------------------------------------------------------- #
# Context manager for signals
HANDLED_SIGNALS = [signal.SIGINT, signal.SIGTERM]


class SignalError(Exception):
    """Error raised by signal handler."""

    pass


def _signal_failure(signum: signal.Signals, frame: FrameType) -> None:
    raise SignalError(str(signum))


class catch_signals(ContextDecorator):
    """Decorator to catch termination signals."""

    def __init__(self: "catch_signals") -> None:
        """Start handler."""
        pass

    def __enter__(self: "catch_signals") -> None:
        """Enter signal handler."""
        self.original_handlers = dict(
            [(s, signal.getsignal(s)) for s in HANDLED_SIGNALS]
        )
        for s in HANDLED_SIGNALS:
            signal.signal(s, _signal_failure)

    def __exit__(self: "catch_signals", exc_type, exc, exc_tb):  # noqa
        """Exit signal handler."""
        for key, val in self.original_handlers.items():
            signal.signal(key, val)


# ----------------------------------------------------------------------------- #
class RunStatus(IntEnum):
    """Possible status of running an operation."""

    # <= 0 -> issue prevented running job.
    subdag_ready = -5
    delayed = -3
    unmet_dependencies = -2
    not_ready = -1
    # > 0 -> executed, can run dependents.
    executed = 1
    using_cached = 2
    input_error = 4


def is_it_cached(db: Redis[bytes], op: Operation) -> bool:
    """Check if an operation is fully cached and doesn't need to be recomputed."""
    # We do this by checking whether all of it's outputs are already saved.
    # This ensures that there is no mismatch between artefact statuses and the
    # status of generating operations.
    keys = [
        join(ARTEFACTS, resolve_link(db, address), "status")
        for address in op.out.values()
    ]

    def __status(p: Redis[bytes]) -> bool:
        for address in op.out.values():
            stat = get_status(p, address)
            if stat <= ArtefactStatus.no_data:
                return False
        return True

    answer: bool = db.transaction(  # type:ignore
        __status, *keys, watch_delay=0.5, value_from_callable=True
    )
    return answer


def dependencies_are_met(db: Redis[bytes], op: Operation) -> bool:
    """Check if all the dependencies of an operation are met."""
    for val in op.inp.values():
        stat = get_status(db, resolve_link(db, val))
        if stat <= ArtefactStatus.no_data:
            return False

    return True


@catch_signals()
def run_op(  # noqa:C901
    db: Redis[bytes],
    store: StorageEngine,
    op: Union[Operation, hash_t],
    evaluate: bool = True,
) -> RunStatus:
    """Run an Operation from its hash address."""
    # Compatibility feature
    if not isinstance(op, Operation):
        op = Operation.grab(db, op)

    logger.info(f"=== {op.hash} ===")
    logger.info("evaluating...")

    # Check if the current job needs to be done at all
    # ------------------------------------------------
    if is_it_cached(db, op):
        # All outputs are ok. We exit this run.
        logger.success("DONE: using cached data.")
        return RunStatus.using_cached

    if not evaluate:
        raise RuntimeError("Attempting to run an operation, but evaluate = False.")

    # # Then we check if all the inputs are ready to be processed.
    if not dependencies_are_met(db, op):
        logger.success("DONE: waiting on dependencies.")
        return RunStatus.unmet_dependencies

    # load the funsie
    funsie = Funsie.grab(db, op.funsie)

    # set options in case we need them in the funsie
    _options_stack.push(op.options)
    # we don't have to pop it later because this process is going to die

    # load input files
    input_data: dict[str, Result[BytesIO]] = {}
    for key, val in op.inp.items():
        artefact = Artefact[Any].grab(db, val)
        dat = get_stream(db, store, artefact.hash, carry_error=op.hash)
        if isinstance(dat, Error):
            if funsie.error_tolerant:
                logger.warning(f"error on input {key} (tolerated).")
                input_data[key] = dat
            else:
                # forward errors and stop
                for val in op.out.values():
                    mark_error(db, val, dat)
                logger.error(f"DONE: error on input {key} (fragile).")
                return RunStatus.input_error
        else:
            input_data[key] = dat

    # Close input bytes when function returns
    def cleanup() -> None:
        for val in input_data.values():
            if isinstance(val, Error):
                pass
            else:
                val.close()

    logger.info("running...")
    try:
        runner = RUNNERS[funsie.how]
        out_data: out_data_t = runner(funsie, input_data)  # type:ignore

    # Timed out
    except rq.timeouts.JobTimeoutException as e:
        for val in op.out.values():
            mark_error(
                db,
                val,
                error=Error(
                    kind=ErrorKind.JobTimedOut,
                    source=op.hash,
                    details=e.args[0],
                ),
            )
        logger.error("DONE: runner timed out.")
        cleanup()
        return RunStatus.executed

    # Killed by signal
    except SignalError as e:
        logger.exception("runner raised!")
        tb_exc = traceback.format_exc()
        # much trouble
        for val in op.out.values():
            mark_error(
                db,
                val,
                error=Error(
                    kind=ErrorKind.KilledBySignal,
                    source=op.hash,
                    details=f"signal={e}",
                ),
            )
        logger.error("DONE: runner killed by signal.")
        cleanup()
        return RunStatus.executed

    # Anything else
    except Exception:
        logger.exception("runner raised!")
        tb_exc = traceback.format_exc()
        # much trouble
        for val in op.out.values():
            mark_error(
                db,
                val,
                error=Error(
                    kind=ErrorKind.ExceptionRaised,
                    source=op.hash,
                    details=tb_exc,
                ),
            )
        logger.error("DONE: runner raised exception.")
        cleanup()
        return RunStatus.executed

    subdag_parents = []
    for key2, val2 in out_data.items():
        if val2 is None:
            logger.warning(f"no output data for {key2}")
            mark_error(
                db,
                op.out[key2],
                error=Error(
                    kind=ErrorKind.MissingOutput,
                    source=op.hash,
                    details="output not returned by runner",
                ),
            )
            # not that in this case, the other outputs are not necesserarily
            # invalidated, only this one.

        elif funsie.how == FunsieHow.subdag:
            # Special case for subdag, create links and update parents
            if isinstance(val2, Artefact):
                create_link(db, op.out[key2], val2.hash)
                subdag_parents += [val2.parent]
            else:
                logger.error("expected artefact, got value")
                mark_error(
                    db,
                    op.out[key2],
                    error=Error(
                        kind=ErrorKind.Mismatch,
                        source=op.hash,
                        details="subdag should have returned an artefact"
                        + f" but it returned {val2}",
                    ),
                )
        else:
            if isinstance(val2, Artefact):
                logger.error(f"expected value, got artefact with hash {val2.hash}")
                mark_error(
                    db,
                    op.out[key2],
                    error=Error(
                        kind=ErrorKind.Mismatch,
                        source=op.hash,
                        details="op should have returned a value"
                        + f" but it returned an artefact {val2}",
                    ),
                )
            else:
                set_data(
                    db,
                    store,
                    op.out[key2],
                    _serdes.encode(funsie.out[key2], val2),
                    status=ArtefactStatus.done,
                )

    if funsie.how == FunsieHow.subdag:
        logger.success("DONE: subdag ready.")
        db.sadd(join(OPERATIONS, op.hash, "parents.subdag"), *subdag_parents)
        logger.info(f"added {len(subdag_parents)} new parent nodes")
        cleanup()
        return RunStatus.subdag_ready
    else:
        logger.success("DONE: successful eval.")
        cleanup()
        return RunStatus.executed



#!/usr/bin/env python
"""Cli utilities."""
from __future__ import annotations

# std
import os
import secrets
import socket
import subprocess
import time
from typing import Optional

# external
import click
from redis import Redis

# module
from . import __version__


def step_arg(step: str, arg: str) -> None:
    """Print a step to output."""
    click.echo(f"{click.style(step, bold=True):30}{arg:62}", nl=False, err=True)


def status_done() -> None:
    """Print DONE."""
    click.echo(click.style("DONE", fg="green", bold=True), err=True)


def status_fail() -> None:
    """Print FAIL."""
    click.echo(click.style("FAIL", fg="red", bold=True), err=True)


def status_skip() -> None:
    """Print FAIL."""
    click.echo(click.style("SKIP", fg="blue", bold=True), err=True)


def info_print(term: str, details: str) -> None:
    """Print formatted funsies hash."""
    click.echo(f"{click.style(term, bold=True):>26} | {details}", err=True)


# This is the main command
@click.command()
@click.version_option(__version__)
@click.argument(
    "run_dir",
    type=click.Path(exists=True),
    default=".",
)
@click.option("--workers", nargs=1, type=int, help="Start INTEGER funsies workers.")
@click.option(
    "--disk/--no-disk",
    default=False,
    help="Save data to disk or within the Redis instance.",
)
@click.option(
    "--pw/--no-pw",
    default=True,
    help="Protect Redis server with a randomly generated password.",
)
@click.option(
    "--port",
    default=6379,
    type=int,
    help="Port for Redis server.",
)
@click.option(
    "--script",
    nargs=1,
    type=click.Path("r"),
    help="Run a python script after initialization.",
)
@click.option(
    "--stop",
    is_flag=True,
    default=False,
    help="Shutdown everything after the script ends.",
)
def main(
    run_dir: Optional[str],
    disk: bool,
    port: int,
    script: Optional[str],
    workers: Optional[int],
    pw: bool,
    stop: bool,
) -> None:
    """Easy initialization of a funsies environment.

    This command condenses the steps required to setup a funsies environment.
    It starts a password protected redis server, sets up the appropriate
    environment variables, initialize the data directory etc.

    To run a full funsies script in one go, do

    start-funsies run_dir --script workflow.py --workers 4 --stop

    In this case the workflow script should have both execute() and wait_for()
    steps, so that it blocks until the work is completed.
    """
    # Set run_directory
    if run_dir is None:
        run_dir = os.getcwd()

    # make global path
    run_dir = os.path.abspath(run_dir)

    # redis parameters
    redis_port = port
    redis_hostname = socket.gethostname()
    if pw:
        redis_password = secrets.token_hex(6)
        redis_url = f"redis://:{redis_password}@{redis_hostname}:{redis_port}"
    else:
        redis_password = ""
        redis_url = f"redis://{redis_hostname}:{redis_port}"

    # data parameters
    if disk:
        data_url = f"file://{os.path.join(run_dir,'funsies_data')}"
    else:
        data_url = redis_url

    # setup the redis settings
    redis_conf = os.path.join(run_dir, "redis.conf")
    env_file = os.path.join(run_dir, "funsies.env")

    info_print("jobs server", redis_url)
    info_print("redis conf", redis_conf)
    info_print("data", data_url)
    info_print("env", env_file)

    step_arg("writing", "redis config")
    with open(redis_conf, "w") as f:
        f.write(f"dir {run_dir}\n")
        f.write("logfile redis.log\n")
        if pw:
            f.write(f"requirepass {redis_password}\n")
        f.write(f"port {redis_port}\n")
        f.write("save 300 1\n")
    status_done()

    # writing env file
    step_arg("writing", "env file")
    with open(env_file, "w") as f:
        f.write(f"export FUNSIES_JOBS='{redis_url}'\n")
        os.environ["FUNSIES_JOBS"] = redis_url
        f.write(f"export FUNSIES_DATA='{data_url}'\n")
        os.environ["FUNSIES_DATA"] = data_url
    status_done()

    # starting the server
    step_arg("redis", "starting the server")
    redis_server = subprocess.Popen(
        ["redis-server", redis_conf],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    time.sleep(0.2)
    stat = redis_server.poll()
    if stat is not None:
        status_fail()
        assert redis_server.stderr is not None
        stdout = redis_server.stderr.read().decode()
        raise RuntimeError(f"Redis server failed to start, errcode={stat}\n{stdout}")
    status_done()

    # wait for server to start
    # timeout = None  # change?
    # if timeout is not None:
    #     tmax = time.time() + timeout

    # wait for the server to load
    step_arg("redis", "connecting to server...")
    while True:
        try:
            db = Redis.from_url(redis_url, decode_responses=False)
            db.ping()
            break
        except Exception:
            time.sleep(1.0)

        # if timeout is not None:
        #     t1 = time.time()
        #     if t1 > tmax:
        #         status_fail()
        #         logger.error("timeout exceeded")
        #         raise SystemExit(2)
    status_done()

    click.echo(
        f"""
Successful initialization!
--------------------------

To complete setup of the funsies environment, source the environment file

        source {env_file}

You should then be able to use funsies commands (funsies --help for a list of
them). You can shutdown the database and worker pool using

        funsies shutdown --all

"""
    )

    if workers is not None:
        if workers > 1:
            s = "s"
        else:
            s = ""
        step_arg("workers", f"starting pool of {workers} worker{s}")
        for i in range(workers):
            subprocess.Popen(
                [
                    "funsies",
                    "--jobs",
                    redis_url,
                    "--data",
                    data_url,
                    "--log",
                    os.path.join(run_dir, f"worker_log.{i}"),
                    "worker",
                ],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            time.sleep(0.1)
        status_done()

    if script is not None:
        step_arg("running script", script)
        out = subprocess.run(
            ["python", script], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if out.returncode != 0:
            status_fail()
            click.echo(f"Warning! error code = {out.returncode}")
        else:
            status_done()
        with open(os.path.join(run_dir, "script.out"), "wb") as fout:
            fout.write(out.stdout)
        with open(os.path.join(run_dir, "script.err"), "wb") as ferr:
            ferr.write(out.stderr)

    if stop:
        step_arg("shutdown", "pool and redis server")
        subprocess.Popen(
            ["funsies", "--jobs", redis_url, "--data", data_url, "shutdown", "--all"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        status_done()


if __name__ == "__main__":
    main()



"""User-facing functions for parametric DAGs."""
from __future__ import annotations

# std
from typing import Any, Optional

# external
from redis import Redis

# module
from ._constants import hash_t
from ._context import get_redis
from ._graph import Artefact
from ._parametrize import make_parametric, Parametric
from .ui import _Target, put


def commit(
    name: str,
    inp: dict[str, Artefact[Any]],
    out: dict[str, Artefact[Any]],
    *,
    connection: Optional[Redis[bytes]] = None,
) -> hash_t:
    """Parametrize and commit a workflow.

    This function gives a name to part of a workflow and parametrizes it for
    use with `parametric.recall`. This is probably best described by an
    example,

    ```
    import funsies as f

    name_john = f.put("john")
    name_John = f.morph(lambda x: x.capitalize(), name)
    f.parametric.commit(
        "capitalize",
        in={"name":name_john},
        out={"name_out":name_John}
    )
    ```

    Now say we wanted to capitalize other names, we can recall the
    `"capitalize"` workflow with new input parameters,

    ```
    out = f.parametric.recall("capitalize", {"name":"tim"})
    f.execute(out["name_out"])
    f.take(out["name_out"])     # returns "Tim"
    ```

    These parametric workflows can be used across scripts, machines, etc. They
    allow encoding steps that can be reproduced without access to the original
    workflow script.

    The graph of task from `inp` to `out` artefacts (and any depedencies) is
    what is saved by `parametric.commit()`.

    Specifically, the `inp` and `out` artefacts defines the workflow to
    encode. Only those input artefacts that will need to be modified on
    `parametric.recall()` need to be explicitly passed. Only those output
    artefacts that we would like computed needs to be passed in `out`. Any
    other necessary inputs will be loaded from the artefact store but not
    parametrized, that is, with the values they had at the time of
    `parametric.commit()`.

    Args:
        name: Name of the parametric DAG.
        inp: Dictionary of artefacts that form the inputs to the Parametric DAG.
        out: Dictionary of artefacts that form the outputs to the Parametric
            DAG. `inp` and `out` together define the committed DAG.
        connection: An explicit Redis connection. Not required if called within a
            `Fun()` context.

    Returns:
        The hash value of the committed Parametric DAG.
    """
    db = get_redis(connection)
    param = make_parametric(db, name, inp, out)
    return param.hash


def recall(
    name_or_hash: str,
    inp: dict[str, _Target],
    *,
    connection: Optional[Redis[bytes]] = None,
) -> dict[str, Artefact[Any]]:
    """Recall a Parametric DAG and evaluate it.

    This function recalls a workflow previously encoded with
    `parametric.commit()`, substitutes in the values in `inp` and returns a
    dictionary of artefacts (from `out` in `parametric.commit()`) that
    descends from the new input values.

    Note that every step is automatically duplicated to account for new data
    if and only if doing so is necessary. That is, the calculation is
    incremental.

    Args:
        name_or_hash: Name or hash value of a DAG encoded with `parametric.commit()`.
        inp: Dictionary of artefacts corresponding to those passed to
            `parametric.commit()`. Any omitted data will simply be substituted by
            whatever was committed.
        connection: An explicit Redis connection. Not required if called within a
            `Fun()` context.

    Returns:
        A dictionary of artefacts that corresponds to the `out=` argument of
        `parametric.commit()`.
    """
    db = get_redis(connection)

    # first, check if its a name
    h = Parametric.resolve_name(db, name_or_hash)
    if h is None:
        # Probably a hash then
        h = hash_t(name_or_hash)

    param = Parametric.grab(db, h)
    new_inps = {}
    for k, arg in inp.items():
        if k not in param.inp:
            raise AttributeError(
                f"input {k} to parametric {h} not in defined inputs"
                + f" {list(param.inp.keys())}"
            )
        if isinstance(arg, Artefact):
            new_inps[k] = arg
        else:
            new_inps[k] = put(arg, connection=db)

    # TODO: encoding check?
    # for k, art in new_inps.items():
    #     if art.kind != param.inp[k].kind:
    #         raise TypeError(
    #             f"input {k} to parametric {h} as encoding {art.kind},"
    #             + f" expected {param.inp[k].kind}"
    #         )

    return param.evaluate(db, new_inps)



#!/usr/bin/env python
"""Cli utilities."""
from __future__ import annotations

# std
import json
import sys
import time
from typing import Optional

# external
import click
from rq import Connection, Worker

# funsies
import funsies

# module
from . import __version__, types  # noqa
from ._constants import hash_t
from ._graph import get_status
from ._logging import logger
from .config import Server

# required funsies libraries loaded in advance
import hashlib, subprocess, loguru  # noqa isort:skip


# This is the main funsies command
@click.group()
@click.version_option(__version__)
@click.option(
    "--jobs",
    "-j",
    type=str,
    nargs=1,
    default=None,
    help="Redis connection URL.",
)
@click.option(
    "--data",
    "-d",
    type=str,
    nargs=1,
    default=None,
    help="Data connection URL.",
)
@click.option(
    "--log",
    type=click.Path(writable=True),
    nargs=1,
    default=None,
    help="Sink output to file.",
)
@click.pass_context
def main(
    ctx: click.Context, jobs: Optional[str], data: Optional[str], log: Optional[str]
) -> None:
    """Command-line tools for funsies.

    The --jobs option allows passing a custom url to locate the redis instance,
    of the form redis://username:password@hostname:port.

    Similarly, the --data options does the same but for data storage. Data
    storage can also be stored on file, using file://$PATH_TO_DATA as the url.

    These options have to come before any of the subcommands (worker, cat
    etc.). They can alternatively be set via the environment variables
    FUNSIES_JOBS and FUNSIES_DATA.
    """
    if log is not None:
        logger.add(log)

    ctx.obj = Server(jobs, data)


@main.command()
@click.option(
    "--burst",
    "-b",
    is_flag=True,
    help="Run in burst mode (quit after all work is done)",
)
@click.option(
    "--rq-log-level",
    type=str,
    default="WARNING",
    help="Set logging for RQ",
    show_default=True,
)
@click.argument("queues", nargs=-1)
@click.pass_context
def worker(ctx: click.Context, queues, burst, rq_log_level):  # noqa:ANN001,ANN201
    """Starts an RQ worker for funsies."""
    db, store = ctx.obj.new_connection()
    funsies._context._storage_stack.push(store)
    with Connection(db):
        queues = queues or ["default"]
        if burst:
            burst_mode = " in burst mode"
        else:
            burst_mode = ""
        logger.success(f"working on queues={', '.join(queues)}{burst_mode}")
        w = Worker(
            queues,
            log_job_description=False,
            # TODO: make adjustable
            serializer="rq.serializers.JSONSerializer",
        )
        w.work(burst=burst, logging_level=rq_log_level)


@main.command()
@click.pass_context
def clean(ctx: click.Context) -> None:
    """Clean job queues and DAGs."""
    db, _ = ctx.obj.new_connection()
    logger.info("shutting down workers")
    funsies._context.shutdown_workers(db, True)
    logger.info("cleaning up")
    funsies._context.cleanup_funsies(db)
    logger.success("done")


@main.command()
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="Shutdown workers without finishing jobs.",
)
@click.option(
    "--all",
    is_flag=True,
    help="Also shutdown job database. Implies --force.",
)
@click.pass_context
def shutdown(ctx: click.Context, force: bool, all: bool) -> None:
    """Tell workers to shutdown."""
    db, _ = ctx.obj.new_connection()
    if all:
        force = True

    funsies._context.shutdown_workers(db, force)
    logger.success("done")

    if all:
        logger.info("shutting down jobs redis instance")
        db.shutdown()
        # TODO what about the data one?


@main.command()
@click.argument("hashes", type=str, nargs=-1)
@click.pass_context
def cat(ctx: click.Context, hashes: tuple[str, ...]) -> None:
    """Print artefacts to stdout."""
    with funsies._context.Fun(ctx.obj):
        for hash in hashes:
            logger.info(f"extracting {hash}")
            things = funsies.get(hash)
            if len(things) == 0:
                logger.error("hash does not correspond to anything!")
                raise SystemExit(2)

            if len(things) > 1:
                logger.error(f"hash resolves to {len(things)} things.")

            art = things[0]
            if isinstance(art, types.Artefact):
                res = funsies.take(art, strict=False)
                if isinstance(res, types.Error):
                    logger.warning(f"error at {hash}: {res.kind}")
                    if res.details is not None:
                        sys.stderr.buffer.write((res.details + "\n").encode())
                    logger.warning(f"error source: {res.source}")
                elif isinstance(res, bytes):
                    sys.stdout.buffer.write(res)
                    logger.success(f"{hash} output to stdout")
                elif isinstance(res, str):
                    sys.stdout.buffer.write(res.encode())
                    logger.success(f"{hash} output to stdout")
                else:
                    sys.stdout.buffer.write(json.dumps(res).encode())
                    logger.success(f"{hash} output to stdout")

            elif isinstance(art, types.Operation):
                logger.error("not an artefact")
                logger.info("did you mean...")
                sys.stderr.write("      INPUTS:\n")
                for key, val in art.inp.items():
                    sys.stderr.write(f"      {key:<30} -> {val[:8]}\n")
                sys.stderr.write("      OUTPUTS:\n")
                for key, val in art.out.items():
                    sys.stderr.write(f"      {key:<30} -> {val[:8]}\n")
            else:
                logger.error("not an artefact:")
                logger.error(f"{art}")


@main.command()
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=False, writable=True),
    help="Output location (defaults to first characters of hash).",
)
@click.argument(
    "hash",
    type=str,
)
@click.pass_context
def debug(ctx: click.Context, hash: str, output: Optional[str]) -> None:
    """Extract all data related to a given hash value."""
    logger.info(f"extracting {hash}")
    if output is None:
        output = hash
        output2 = ""
    else:
        output2 = output + "_"

    with funsies._context.Fun(ctx.obj):
        things = funsies.get(hash)
        if len(things) == 0:
            logger.error(f"{hash} does not correspond to anything!")
        elif len(things) == 1:
            logger.info(f"got {type(things[0])}")
            funsies.debug.anything(things[0], output)
            logger.success(f"saved to {output}")

        else:
            logger.warning(f"got {len(things)} objects")
            funsies.debug.anything(things[0], output)
            logger.success(f"{type(things[0])} -> {output}")
            for el in things[1:]:
                funsies.debug.anything(el, output2 + el.hash)
                logger.success(f"{type(el)} -> {output2 + el.hash}")


@main.command()
@click.argument(
    "hashes",
    type=str,
    nargs=-1,
)
@click.pass_context
def reset(ctx: click.Context, hashes: tuple[str, ...]) -> None:
    """Reset operations and their dependents."""
    with funsies._context.Fun(ctx.obj):
        for hash in hashes:
            things = funsies.get(hash)
            if len(things) == 0:
                logger.warning(f"no object with hash {hash}")
            if len(things) > 1:
                logger.error(f"more than one object with hash starting with {hash}")
                logger.info("which one do you mean of :")
                for t in things:
                    logger.info(f"\t{t.hash}")
                logger.info("none were reset")
            else:
                if isinstance(things[0], types.Artefact) or isinstance(
                    things[0], types.Operation
                ):
                    funsies.ui.reset(things[0])
                    logger.success(f"{hash} was reset")
                else:
                    logger.error(
                        f"object {hash} is neither an operation or an artefact"
                    )
                    logger.info(f"{hash} was not reset")


@main.command()
@click.argument(
    "hashes",
    type=str,
    nargs=-1,
)
@click.option("-t", "--timeout", type=float, help="Timeout in seconds.")
@click.pass_context
def wait(  # noqa:C901
    ctx: click.Context, hashes: tuple[str, ...], timeout: Optional[float]
) -> None:
    """Wait until redis database or certain hashes are ready."""
    if timeout is not None:
        tmax = time.time() + timeout

    while True:
        try:
            db, _ = ctx.obj.new_connection(try_fail=True)
            db.ping()
            break
        except Exception:
            time.sleep(1.0)

        if timeout is not None:
            t1 = time.time()
            if t1 > tmax:
                logger.error("timeout exceeded")
                raise SystemExit(2)

    with funsies._context.Fun(ctx.obj):
        db = funsies._context.get_redis()
        h = []
        for hash in hashes:
            things = funsies.get(hash)
            if len(things) == 0:
                logger.warning(f"no object with hash {hash}")
            for t in things:
                if isinstance(t, types.Artefact):
                    h += [t.hash]
                    logger.info(f"waiting on artefact at {hash}")
                elif isinstance(t, types.Operation):
                    h += [next(iter(t.out.values()))]
                    logger.info(f"waiting on operation at {hash}")
                else:
                    logger.warning(f"ignoring {type(t)} at {t.hash}")

        while len(h) > 0:
            stat = get_status(db, h[0], resolve_links=True)
            if stat > 0:
                h.pop(0)
                logger.success(f"{len(h)} things left to wait for.")
            time.sleep(0.5)

            if timeout is not None:
                t1 = time.time()
                if t1 > tmax:
                    logger.error("timeout exceeded")
                    raise SystemExit(2)


@main.command()
@click.argument(
    "hashes",
    type=str,
    nargs=-1,
)
@click.option(
    "-i",
    "--inputs",
    is_flag=True,
)
@click.pass_context
def graph(ctx: click.Context, hashes: tuple[str, ...], inputs: bool) -> None:
    """Print to stdout a DOT-formatted graph to visualize DAGs."""
    # funsies
    import funsies._graphviz

    with funsies._context.Fun(ctx.obj):
        db = funsies._context.get_redis()
        if len(hashes) == 0:
            # If no hashes are passed, we graph all the DAGs on index
            hashes = tuple(
                [dag.decode() for dag in db.smembers(funsies._constants.DAG_INDEX)]
            )

        all_data: list[hash_t] = []
        for hash in hashes:
            things = funsies.get(hash.split("/")[-1])
            if len(things) == 0:
                logger.warning(f"no object with hash {hash}")
            for t in things:
                if isinstance(t, types.Operation) or isinstance(t, types.Artefact):
                    all_data += [hash_t(t.hash)]

        if len(all_data):
            logger.info(f"writing graph for {len(all_data)} objects")
            out = funsies._graphviz.format_dot(
                *funsies._graphviz.export(db, all_data),
                targets=all_data,
                show_inputs=inputs,
            )
            sys.stdout.write(out)
            logger.success("done")
        else:
            logger.error("No data points")
            raise SystemExit(2)


@main.command()
@click.argument(
    "hashes",
    type=str,
    nargs=-1,
)
@click.pass_context
def execute(ctx: click.Context, hashes: tuple[str, ...]) -> None:
    """Enqueue execution of hashes."""
    with funsies._context.Fun(ctx.obj):
        exec_list = []
        for hash in hashes:
            things = funsies.get(hash)
            if len(things) == 0:
                logger.warning(f"no object with hash {hash}")
            for t in things:
                if isinstance(t, types.Operation) or isinstance(t, types.Artefact):
                    exec_list += [t]
                else:
                    logger.warning(f"object with hash {hash} of type {type(t)} skipped")

        funsies.execute(*exec_list)


if __name__ == "__main__":
    main()



"""Contains `template()`, a function to create templated scripts."""
from __future__ import annotations

# std
import os
from typing import Any, Mapping, Optional, TypeVar, Union

# external
import chevron
from redis import Redis

# module
from ._constants import Encoding
from ._context import Connection, get_connection, get_options
from ._graph import Artefact, constant_artefact, make_op
from ._pyfunc import python_funsie
from ._storage import StorageEngine
from .config import Options

T = TypeVar("T")


def _artefact(
    db: Redis[bytes], store: StorageEngine, data: Union[T, Artefact[T]]
) -> Artefact[T]:
    if isinstance(data, Artefact):
        return data
    else:
        return constant_artefact(db, store, data)  # type:ignore


_Template = Union[str, Artefact[str], bytes, Artefact[bytes]]
_Value = Any


def template(
    template: _Template,
    data: Mapping[str, _Value],
    strip: bool = True,
    *,
    env: Optional[Mapping[str, str]] = None,
    name: Optional[str] = None,
    opt: Optional[Options] = None,
    connection: Connection = None,
) -> Artefact[bytes]:
    """Fill in a template using data from artefacts.

    This function takes a [chevron template](https://mustache.github.io/) and
    fills it with the dictionary ``data``,

    ```python
    funsies = f.template(template, data)
    # corresponds basically to running
    normal = chevron.render(template, data)
    ```

    ``template()`` is a full-featured funsies function: both the template and
    the data can come from the database and are (as usual) lazily evaluated.
    Substitutions provided in the ``env=`` dictionary are expanded using
    environment variables.

    The primary intended use of ``template()`` is the generation of input
    files for simulation software. For example,

    ```python
    # funsies
    import funsies as f

    g16_template = \"""%NProcShared={{nthreads}}
    # {{functional}}/{{basis}} Symm=None {{type}}

    Gaussian calculation

    {{charge}} {{spin}}
    {{structure}}

    \"""

    with f.Fun():
        inp = f.template(
            g16_template,
            {
                "functional": "b3lyp",
                "basis": "6-31g",
                "type": "sp",
                "spin": 1,
                # the next two could be obtained eg from conformer generation.
                "charge": charge,
                "structure": coords,
            },
            env={"nthreads": "OMP_NUM_THREADS"},
        )
        dft_job = f.shell(
            "g16 input.com", inp={"input.com": inp}, out=["input.log", "data.chk"]
        )
    ```

    Args:
        template: The template, either as a string or as an ``types.Artefact[str]``.
        data: A ``dict[key, value]`` of substitutions to perform on the
            template. ``value`` can be any type accepted by ``chevrons``
            (``str`` but also ``int``, ``bytes`` etc.) and/or ``types.Artefact``
            objects containing those types.
        strip: If `True`, substitutions will be ``.strip()`` before templating.
        env: A ``dict[str,str]`` of substitutions to fill in from the
            environment variables of the worker process.
        name: Provide an explicit name to the template.
        connection: An explicit Redis connection. Not required if called within a
            `Fun()` context.
        opt: An `types.Options` instance as returned by `options()`. Not
            required if called within a `Fun()` context.

    Returns:
        An `types.Artefact[bytes]` object, populated with the generated
        template as a bytestring.
    """  # noqa:D300,D301
    # Get context elements
    opt = get_options(opt)
    db, store = get_connection(connection)

    # Make sure template is a string in the db
    if isinstance(template, bytes):
        tmp = _artefact(db, store, template.decode())
    else:
        tmp = _artefact(db, store, template)  # type:ignore

    # Make sure all substitutions are strings in the db
    args = dict()
    for key, value in data.items():
        if isinstance(value, bytes):
            value = value.decode()
        args[key] = _artefact(db, store, value)

    template_key = "template"
    while template_key in args:
        template_key += "_"  # append more _ until template_key is unique

    env_key = "env"
    while env_key in args:
        env_key += "_"  # append more _ until template_key is unique

    args[template_key] = tmp
    args[env_key] = _artefact(db, store, env)

    in_types = dict([(k, val.kind) for k, val in args.items()])

    if name is not None:
        fun_name = name
    else:
        do_strip = ""
        if strip:
            do_strip = ", stripped"
        fun_name = f"template (chevrons{do_strip})"

    def __exec(inpd: Mapping[str, Any]) -> dict[str, bytes]:
        """Substitute into template."""
        args = {}
        for key, val in inpd.items():
            if isinstance(val, bytes):
                val = val.decode()
            if isinstance(val, str) and strip and key != template_key:
                val = val.strip()
            args[key] = val

        # read template
        template = args[template_key]

        # read env variables
        if args[env_key] is not None:
            env = args[env_key]
            for key, val in env.items():
                args[key] = os.environ.get(val)

        del args[template_key]
        del args[env_key]

        return {"out": chevron.render(template, args).encode()}

    funsie = python_funsie(
        __exec, in_types, {"out": Encoding.blob}, name=fun_name, strict=True
    )
    operation = make_op(db, funsie, args, opt)
    return Artefact.grab(db, operation.out["out"])



"""Helpful function for debugging workflows."""
from __future__ import annotations

# std
from dataclasses import asdict
import json
import os
import os.path
from typing import Any, Union

# module
from ._constants import _AnyPath
from ._context import Connection, get_connection
from ._funsies import Funsie, FunsieHow
from ._graph import Artefact, get_data, Operation
from ._shell import ShellOutput
from .errors import UnwrapError
from .ui import takeout


# ----------------------------------------------------------------------
# Debugging functions
def shell(  # noqa:C901
    shell_output: ShellOutput,
    directory: _AnyPath,
    connection: Connection = None,
) -> None:
    """Extract all the files and outputs of a shell function to a directory."""
    os.makedirs(directory, exist_ok=True)
    inp = os.path.join(directory, "input_files")
    out = os.path.join(directory, "output_files")
    db, store = get_connection(connection)
    errors = {}

    for key, val in shell_output.inp.items():
        try:
            p = os.path.join(inp, key)
            os.makedirs(os.path.dirname(p), exist_ok=True)
            takeout(val, p, connection=(db, store))
        except UnwrapError:
            errors[f"input:{key}"] = asdict(get_data(db, store, val))

    for key, val in shell_output.out.items():
        try:
            p = os.path.join(out, key)
            os.makedirs(os.path.dirname(p), exist_ok=True)
            takeout(val, p, connection=(db, store))
        except UnwrapError:
            errors[f"output:{key}"] = asdict(get_data(db, store, val))

    for i in range(len(shell_output.stdouts)):
        try:
            takeout(
                shell_output.stdouts[i],
                os.path.join(directory, f"stdout{i}"),
                connection=(db, store),
            )
        except UnwrapError:
            errors[f"stdout:{key}"] = asdict(get_data(db, store, val))

        try:
            takeout(
                shell_output.stderrs[i],
                os.path.join(directory, f"stderr{i}"),
                connection=(db, store),
            )
        except UnwrapError:
            errors[f"stderr:{key}"] = asdict(get_data(db, store, val))

    with open(os.path.join(directory, "errors.json"), "w") as f:
        f.write(
            json.dumps(
                errors,
                sort_keys=True,
                indent=2,
            )
        )

    with open(os.path.join(directory, "operation.json"), "w") as f:
        f.write(json.dumps(asdict(shell_output.op), sort_keys=True, indent=2))

    extra = Funsie.grab(db, shell_output.op.funsie).extra
    cmds = json.loads(extra["cmds"].decode())
    env = json.loads(extra["env"].decode())
    with open(os.path.join(directory, "op.sh"), "w") as f:
        f.write("\n".join(cmds))
        f.write("\n")

    if env is not None:
        with open(os.path.join(directory, "op.env"), "w") as f:
            for key, val in env.items():
                f.write(f"{key}={val}\n")


# --------------
# Debug artefact
def artefact(
    target: Artefact,
    directory: _AnyPath,
    connection: Connection = None,
) -> None:
    """Output content of any hash object to a file."""
    db, store = get_connection(connection)
    os.makedirs(directory, exist_ok=True)
    with open(os.path.join(directory, "metadata.json"), "w") as f:
        f.write(json.dumps(asdict(target), sort_keys=True, indent=2))
    try:
        takeout(
            target,
            os.path.join(directory, "data"),
            connection=(db, store),
        )
    except UnwrapError:
        # dump error to json file
        with open(os.path.join(directory, "error.json"), "w") as f:
            f.write(
                json.dumps(
                    asdict(get_data(db, store, target)),
                    sort_keys=True,
                    indent=2,
                )
            )


def python(
    target: Union[Operation, Artefact],
    directory: _AnyPath,
    connection: Connection = None,
) -> None:
    """Output content of any hash object to a file."""
    db, store = get_connection(connection)

    if isinstance(target, Artefact):
        # Get the corresponding operation
        target = Operation.grab(db, target.parent)
        if target is None:
            raise RuntimeError(f"Operation not found at {target.parent}")

    os.makedirs(directory, exist_ok=True)
    funsie = Funsie.grab(db, target.funsie)
    inp = os.path.join(directory, "inputs")
    out = os.path.join(directory, "outputs")
    errors = {}
    if funsie.how != FunsieHow.python:
        raise RuntimeError(f"Operation is of type {funsie.how}, not a python function.")

    for key, v in target.inp.items():
        val = Artefact[Any].grab(db, v)
        try:
            p = os.path.join(inp, key)
            os.makedirs(os.path.dirname(p), exist_ok=True)
            takeout(val, p, connection=(db, store))
        except UnwrapError:
            errors[f"input:{key}"] = asdict(get_data(db, store, val))

    for key, v in target.out.items():
        val = Artefact.grab(db, v)
        try:
            p = os.path.join(out, key)
            os.makedirs(os.path.dirname(p), exist_ok=True)
            takeout(val, p, connection=(db, store))
        except UnwrapError:
            errors[f"output:{key}"] = asdict(get_data(db, store, val))

    with open(os.path.join(directory, "errors.json"), "w") as f:
        f.write(
            json.dumps(
                errors,
                sort_keys=True,
                indent=2,
            )
        )

    meta = {
        "what": funsie.what,
        "inp": funsie.inp,
        "out": funsie.out,
        "error_tolerant": funsie.error_tolerant,
    }
    with open(os.path.join(directory, "funsie.json"), "w") as f:
        f.write(json.dumps(meta, sort_keys=True, indent=2))

    with open(os.path.join(directory, "operation.json"), "w") as f:
        f.write(json.dumps(asdict(target), sort_keys=True, indent=2))

    # TODO
    # with open(os.path.join(directory, "function.pkl"), "wb") as f:
    #     assert funsie.aux is not None
    #     f.write(funsie.aux)


# --------------
# Debug anything
def anything(
    obj: Union[Artefact, Funsie, Operation, ShellOutput],
    output: _AnyPath,
    connection: Connection = None,
) -> None:
    """Debug anything really."""
    db, store = get_connection(connection)
    if isinstance(obj, Operation):
        funsie = Funsie.grab(db, obj.funsie)
        if funsie.how == FunsieHow.shell:
            shell_output = ShellOutput(db, obj)
            shell(shell_output, output, connection=(db, store))
        elif funsie.how == FunsieHow.python:
            python(obj, output, connection=(db, store))
        else:
            raise RuntimeError()
    elif isinstance(obj, Artefact):
        artefact(obj, output, connection=(db, store))
    elif isinstance(obj, ShellOutput):
        shell(obj, output, connection=(db, store))
    else:
        raise NotImplementedError(f"Object of type {obj} cannot be debugged.")



"""Functions for describing redis-backed DAGs."""
from __future__ import annotations

# std
from dataclasses import dataclass
from enum import IntEnum
import hashlib
import io
from typing import Any, Generic, Mapping, Optional, Type, TypeVar

# external
from redis import Redis
from redis.client import Pipeline

# module
from . import _serdes
from ._constants import _Data, ARTEFACTS, Encoding, hash_t, join, OPERATIONS
from ._funsies import Funsie
from ._logging import logger
from ._short_hash import hash_save
from ._storage import descr_t, StorageEngine
from .config import Options
from .errors import Error, ErrorKind, match, Result

# Max redis value size in bytes
MIB = 1024 * 1024
MAX_VALUE_SIZE = 512 * MIB


# --------------------------------------------------------------------------------
# Artefact status
class ArtefactStatus(IntEnum):
    """Status of data associated with an artefact."""

    deleted = -2
    not_found = -1
    no_data = 0
    # > absent  =  artefact has been computed
    done = 1
    const = 2
    error = 3
    linked = 4


def get_status(
    db: Redis[bytes], address: hash_t, resolve_links: bool = False
) -> ArtefactStatus:
    """Get the status of an artefact."""
    if resolve_links:
        address = resolve_link(db, address)

    val = db.get(join(ARTEFACTS, address, "status"))
    if val is None:
        return ArtefactStatus.not_found
    else:
        return ArtefactStatus(int(val))


def set_status(db: Redis[bytes], address: hash_t, stat: ArtefactStatus) -> None:
    """Set the status of an artefact."""
    _ = db.set(join(ARTEFACTS, address, "status"), int(stat))


def set_status_nx(db: Redis[bytes], address: hash_t, stat: ArtefactStatus) -> None:
    """Set the status of an artefact iff it has no status."""
    _ = db.setnx(join(ARTEFACTS, address, "status"), int(stat))


# --------------------------------------------------------------------------------
# Generic Artefacts
T = TypeVar("T")
Tdata = TypeVar("Tdata", bound=_Data)


@dataclass(frozen=True)
class Artefact(Generic[T]):
    """Artefacts are the main data structure."""

    hash: hash_t
    parent: hash_t
    kind: Encoding

    def put(self: Artefact[Any], db: Redis[bytes]) -> None:
        """Save an artefact to Redis."""
        data = dict(hash=self.hash, parent=self.parent, kind=self.kind.value)
        db.hset(
            join(ARTEFACTS, self.hash),
            mapping=data,  # type:ignore
        )
        # Save the hash in the quickhash db
        hash_save(db, self.hash)

    @classmethod
    def grab(cls: Type[Artefact[T]], db: Redis[bytes], hash: hash_t) -> Artefact[T]:
        """Grab an artefact from the Redis store."""
        pipe: Pipeline = db.pipeline(transaction=False)
        pipe.exists(join(ARTEFACTS, hash))
        pipe.hgetall(join(ARTEFACTS, hash))
        exists, data = pipe.execute()

        if not exists:
            raise RuntimeError(f"No artefact at {hash}")

        return Artefact[T](
            hash=hash_t(data[b"hash"].decode()),
            parent=hash_t(data[b"parent"].decode()),
            kind=Encoding(data[b"kind"].decode()),
        )


def is_artefact(db: Redis[bytes], address: hash_t) -> bool:
    """Check whether a hash corresponds to an artefact."""
    return bool(db.exists(join(ARTEFACTS, address)))


# Mark artefacts
def mark_done(db: Redis[bytes], address: hash_t) -> None:
    """Set the status of a given operation or artefact."""
    assert is_artefact(db, address)
    old = get_status(db, address)
    if old == ArtefactStatus.const:
        logger.error("attempted to mark done a const artefact.")
    else:
        set_status(db, address, ArtefactStatus.done)


def mark_error(db: Redis[bytes], address: hash_t, error: Error) -> None:
    """Set the status of a given operation or artefact."""
    assert is_artefact(db, address)
    old = get_status(db, address)
    if old == ArtefactStatus.const:
        logger.error("attempted to mark in error a const artefact.")
    else:
        set_status(db, address, ArtefactStatus.error)
        error.put(db, address)


# Delete artefact
def delete_artefact(db: Redis[bytes], address: hash_t) -> None:
    """Set the status of a given operation or artefact."""
    assert is_artefact(db, address)
    old = get_status(db, address)
    if old == ArtefactStatus.const:
        logger.error(f"attempted to delete const artefact {address[:6]}.")
        logger.error("you should set it to a different value instead")
        return

    # mark as deleted
    set_status(db, address, ArtefactStatus.deleted)

    # delete data
    db.delete(join(ARTEFACTS, address, "data"))


# Make an artefact point to another artefact
def create_link(db: Redis[bytes], afrom: hash_t, ato: hash_t) -> None:
    """Set the status of a given operation or artefact."""
    # TODO do not link to artefact of different type!
    assert is_artefact(db, afrom)
    assert is_artefact(db, ato)
    old = get_status(db, afrom)
    if old == ArtefactStatus.const:
        logger.error("attempted to link a const artefact to something else.")
    else:
        set_status(db, afrom, ArtefactStatus.linked)
        key = join(ARTEFACTS, afrom, "links_to")
        db.set(key, ato)


def resolve_link(db: Redis[bytes], address: hash_t) -> hash_t:
    """Resolve any link recursively."""
    key = join(ARTEFACTS, address, "links_to")
    link = db.get(key)
    if link is None:
        return address
    else:
        out = hash_t(link.decode())
        return resolve_link(db, out)


def __get_data_loc(
    db: Redis[bytes],
    store: StorageEngine,
    address: hash_t,
    carry_error: Optional[hash_t] = None,
    do_resolve_link: bool = True,
) -> Result[descr_t]:
    """Perform all the prior step before actually retrieving data."""
    if do_resolve_link:
        address = resolve_link(db, address)

    # First check the status
    stat = get_status(db, address)

    # if it's a link, we move over to the link
    if stat == ArtefactStatus.linked:
        return Error(
            kind=ErrorKind.UnresolvedLink,
            details=f"artefact at {address} is a link and thus has no data",
            source=carry_error,
        )
    elif stat == ArtefactStatus.error:
        return Error.grab(db, address)
    elif stat <= ArtefactStatus.no_data:
        return Error(
            kind=ErrorKind.NotFound,
            details=f"No data associated with artefact: {stat}.",
            source=carry_error,
        )
    else:
        key = store.get_key(address)
        return key


# Save and load artefacts
def get_stream(
    db: Redis[bytes],
    store: StorageEngine,
    source: hash_t,
    carry_error: Optional[hash_t] = None,
    do_resolve_link: bool = True,
) -> Result[io.BytesIO]:
    """Retrieve data corresponding to an artefact."""
    key = __get_data_loc(db, store, source, carry_error, do_resolve_link)
    if isinstance(key, Error):
        return key

    stream = store.take(key)

    if isinstance(stream, Error):
        return Error(
            kind=stream.kind,
            details=stream.details,
            source=carry_error,
        )

    return stream


def get_data(
    db: Redis[bytes],
    store: StorageEngine,
    source: Artefact[T],
    carry_error: Optional[hash_t] = None,
    do_resolve_link: bool = True,
) -> Result[T]:
    """Retrieve data corresponding to an artefact."""
    stream = get_stream(db, store, source.hash, carry_error, do_resolve_link)
    if isinstance(stream, Error):
        return stream
    else:
        raw = stream.read()
        stream.close()
        return _serdes.decode(source.kind, raw, carry_error=carry_error)


def set_stream(
    db: Redis[bytes],
    store: StorageEngine,
    address: hash_t,
    value: Result[io.BytesIO],
    status: ArtefactStatus,
) -> None:
    """Update an artefact with a stream of bytes."""
    if get_status(db, address) == ArtefactStatus.const:
        if status == ArtefactStatus.const:
            pass
        else:
            raise TypeError("Attempted to set data to a const artefact.")

    if isinstance(value, Error):
        # fail gracefully
        mark_error(db, address, error=value)
        return

    key = store.get_key(address)
    stat = store.put(key, value)

    if stat is not None:
        mark_error(db, address, error=stat)
    else:
        set_status(db, address, status)


def set_data(
    db: Redis[bytes],
    store: StorageEngine,
    address: hash_t,
    value: Result[bytes],
    status: ArtefactStatus,
) -> None:
    """Update an artefact with a value."""
    # This function is just a wrapper around set_stream.
    buf = match(value, lambda x: io.BytesIO(x), lambda x: x)
    set_stream(db, store, address, buf, status)


# not pipeline-able (because of set_data)
def constant_artefact(
    db: Redis[bytes], store: StorageEngine, value: Tdata
) -> Artefact[Tdata]:
    """Db an artefact with a defined value."""
    kind = _serdes.kind(value)
    data = _serdes.encode(kind, value)
    if isinstance(data, Error):
        raise TypeError(f"constant artefact could not be encoded:\n{data}")

    # ==============================================================
    #     ALERT: DO NOT TOUCH THIS CODE WITHOUT CAREFUL THOUGHT
    # --------------------------------------------------------------
    # When hashes change, previous databases become deprecated. This
    # (will) require a change in version number!
    m = hashlib.sha1()
    m.update(b"artefact\n")
    m.update(b"constant\n")
    m.update(f"kind:{str(kind)}\n".encode())
    m.update(data)
    h = hash_t(m.hexdigest())
    # ==============================================================
    node = Artefact[Tdata](hash=h, parent=hash_t("root"), kind=kind)
    pipe: Pipeline = db.pipeline(transaction=False)
    node.put(pipe)
    pipe.execute()
    set_data(db, store, h, data, status=ArtefactStatus.const)
    return node


# pipeline-able
def variable_artefact(
    store: Redis[bytes],
    parent_hash: hash_t,
    name: str,
    kind: Encoding,
) -> Artefact[Any]:
    """Store an artefact with a generated value."""
    # ==============================================================
    #     ALERT: DO NOT TOUCH THIS CODE WITHOUT CAREFUL THOUGHT
    # --------------------------------------------------------------
    # When hashes change, previous databases become deprecated. This
    # (will) require a change in version number!
    m = hashlib.sha1()
    m.update(b"artefact\n")
    m.update(b"variable\n")
    m.update(f"kind:{str(kind)}\n".encode())
    m.update(f"parent:{parent_hash}\n".encode())
    m.update(f"name:{name}\n".encode())
    h = hash_t(m.hexdigest())
    # ==============================================================
    node = Artefact[T](hash=h, parent=parent_hash, kind=kind)
    node.put(store)
    set_status_nx(store, node.hash, ArtefactStatus.no_data)
    return node


# --------------------------------------------------------------------------------
# Operations
@dataclass(frozen=True)
class Operation:
    """An operation on data in the graph."""

    hash: hash_t
    funsie: hash_t
    inp: dict[str, hash_t]
    out: dict[str, hash_t]
    options: Optional[Options] = None

    # pipeline-able
    def put(self: "Operation", db: Redis[bytes]) -> None:
        """Save an operation to Redis."""
        if self.inp:
            db.hset(join(OPERATIONS, self.hash, "inp"), mapping=self.inp)  # type:ignore
        if self.out:
            db.hset(join(OPERATIONS, self.hash, "out"), mapping=self.out)  # type:ignore
        if self.options:
            db.set(join(OPERATIONS, self.hash, "options"), self.options.pack())
        db.hset(
            join(OPERATIONS, self.hash),
            mapping={"funsie": self.funsie, "hash": self.hash},
        )

        # Save the hash in the quickhash db
        hash_save(db, self.hash)

    # pipelined
    @classmethod
    def grab(cls: Type["Operation"], db: Redis[bytes], hash: hash_t) -> "Operation":
        """Grab an operation from the Redis store."""
        if not db.exists(join(OPERATIONS, hash)):
            raise RuntimeError(f"No operation at {hash}")

        pipe: Pipeline = db.pipeline(transaction=False)
        pipe.hgetall(join(OPERATIONS, hash))
        pipe.hgetall(join(OPERATIONS, hash, "inp"))
        pipe.hgetall(join(OPERATIONS, hash, "out"))
        pipe.get(join(OPERATIONS, hash, "options"))
        metadata, inp, out, tmp = pipe.execute()

        if tmp is not None:
            options: Optional[Options] = Options.unpack(tmp.decode())
        else:
            options = None

        return Operation(
            hash=hash_t(metadata[b"hash"].decode()),
            funsie=hash_t(metadata[b"funsie"].decode()),
            inp=dict([(k.decode(), hash_t(v.decode())) for k, v in inp.items()]),
            out=dict([(k.decode(), hash_t(v.decode())) for k, v in out.items()]),
            options=options,
        )


# pipelined
def make_op(
    store: Redis[bytes], funsie: Funsie, inp: Mapping[str, Artefact[Any]], opt: Options
) -> Operation:
    """Store an artefact with a defined value."""
    # Setup the input artefacts.
    inp_art = {}
    for key in inp:
        if key not in funsie.inp:
            raise AttributeError(f"Extra key {key} passed to funsie.")
        if inp[key].kind != funsie.inp[key]:
            raise TypeError(
                f"Key {key} has type {inp[key].kind} but funsie"
                + f" expects {funsie.inp[key]}"
            )

    for key in funsie.inp:
        if key not in inp:
            raise AttributeError(f"Missing {key} from inputs required by funsie.")
        else:
            inp_art[key] = inp[key].hash

    # Setup a buffered command pipeline for performance
    pipe: Pipeline = store.pipeline(transaction=False)

    # save funsie
    funsie.put(pipe)

    # ==============================================================
    #     ALERT: DO NOT TOUCH THIS CODE WITHOUT CAREFUL THOUGHT
    # --------------------------------------------------------------
    # When hashes change, previous databases become deprecated. This
    # (will) require a change in version number!
    m = hashlib.sha1()
    m.update(b"op")
    m.update(funsie.hash.encode())
    for key, val in sorted(inp.items(), key=lambda x: x[0]):
        m.update(f"file={key}, hash={val}".encode())
    ophash = hash_t(m.hexdigest())
    # ==============================================================

    # Setup the output artefacts.
    out_art = {}
    for key, type in funsie.out.items():
        out_art[key] = variable_artefact(pipe, ophash, key, type).hash

    # Make the node
    node = Operation(ophash, funsie.hash, inp_art, out_art, opt)

    # store the node
    node.put(pipe)

    # Add parents
    root = True
    for k in inp_art.keys():
        v = inp[k]
        # register op as dependent of artefacts
        pipe.sadd(join(ARTEFACTS, v.hash, "dependents"), ophash)

        if v.parent != "root":
            # register ops as dependents of other ops
            pipe.sadd(join(OPERATIONS, ophash, "parents"), v.parent)
            pipe.sadd(join(OPERATIONS, v.parent, "children"), ophash)
            root = False
    if root:
        # This dag has no dependencies
        pipe.sadd(join(OPERATIONS, ophash, "parents"), "root")
        pipe.sadd(join(OPERATIONS, hash_t("root"), "children"), ophash)

    # Execute the full transaction
    pipe.execute()
    return node


def get_op_options(store: Redis[bytes], hash: hash_t) -> Options:
    """Load an operation from Redis store."""
    # store the artefact
    out = store.get(join(OPERATIONS, hash, "options"))
    if out is None:
        raise RuntimeError(f"Options for operation at {hash} could not be found.")
    return Options.unpack(out.decode())



"""Infer function types."""
from __future__ import annotations

# std
from typing import Any, Callable, get_type_hints, Tuple, Type, Union

# python 3.7 imports Literal from typing_extensions
try:
    # std
    from typing import get_args, get_origin
except ImportError:
    from typing_extensions import get_args, get_origin

# module
from ._constants import Encoding
from ._logging import logger


def __translate(x: Union[Type[Any], str]) -> Encoding:
    if x is bytes or x == "bytes":
        return Encoding.blob
    else:
        return Encoding.json


# Type a function
def output_types(fun: Callable[..., Any]) -> tuple[Encoding, ...]:
    """Infer number of return values from a function's type hint."""
    try:
        hints = get_type_hints(fun)
    except NameError:
        hints = fun.__annotations__
        logger.warning(
            "type inference met an unknown reference\n"
            + f"signature: {hints} name: {fun.__name__}"
        )

    if "return" not in hints:
        if fun.__name__ == "<lambda>":
            logger.debug("warning: lambda function has no return types")

        raise TypeError(
            "Failed to infer output types: return annotation is absent.\n"
            + f"signature: {hints} name: {fun.__name__}\n"
            + "You need to define your outputs explicitly."
        )
    else:
        returns = hints["return"]

    if type(returns) == str:
        # TODO attempt to interpret it

        raise TypeError(
            "Failed to infer output types: return annotation is absent.\n"
            + f"signature: {hints} name: {fun.__name__}\n"
            + "Return value converted to string.\nAre you using"
            + "from __future__ import annotations on python=3.7?"
        )
    else:
        generic = get_origin(returns)
        if (
            generic is tuple
            or generic is Tuple
            or generic == "Tuple"
            or generic == "tuple"
        ):
            args = get_args(returns)
            if Ellipsis in args:
                raise TypeError(
                    "Failed to infer output types: return annotation is too broad.\n"
                    + f"signature: {hints} name: {fun.__name__}\n"
                    + "It's not clear how many elements this function should output."
                    + "You need to define your outputs explicitly"
                )
            return tuple(__translate(t) for t in args)
        else:
            return (__translate(returns),)



"""Object types."""
# module
from ._constants import Encoding, hash_t
from ._funsies import Funsie, FunsieHow
from ._graph import Artefact, ArtefactStatus, Operation
from ._run import RunStatus
from ._shell import ShellOutput
from .config import Options
from .errors import Error, ErrorKind, Result, UnwrapError

# A simple mypy result type
Result = Result
"""see `funsies.errors.Result`."""

__all__ = [
    "Funsie",
    "FunsieHow",
    "Artefact",
    "ArtefactStatus",
    "Encoding",
    "Operation",
    "ShellOutput",
    "hash_t",
    "Error",
    "ErrorKind",
    "Result",
    "UnwrapError",
    "RunStatus",
    "Options",
]



"""User-friendly interfaces to funsies functionality."""
from __future__ import annotations

# std
import shutil
import time
from typing import Iterable, Mapping, Optional, overload, TypeVar, Union

# external
from redis import Redis

# python 3.7 imports Literal from typing_extensions
try:
    # std
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type:ignore

# module
from ._constants import _AnyPath, _Data, hash_t
from ._context import Connection, get_connection, get_options
from ._dag import descendants, start_dag_execution
from ._graph import (
    Artefact,
    constant_artefact,
    delete_artefact,
    get_data,
    get_status,
    get_stream,
    make_op,
    Operation,
)
from ._logging import logger
from ._run import is_it_cached
from ._shell import shell_funsie, ShellOutput
from ._short_hash import shorten_hash
from ._storage import StorageEngine
from .config import Options
from .errors import Error, Result, unwrap

# Types
_Target = Union[Artefact, _Data]
_INP_FILES = Optional[Mapping[_AnyPath, _Target]]
_OUT_FILES = Optional[Iterable[_AnyPath]]
T = TypeVar("T", bound=_Data)


def _artefact(
    db: Redis[bytes], store: StorageEngine, data: Union[T, Artefact[T]]
) -> Artefact[T]:
    if isinstance(data, Artefact):
        return data
    else:
        return constant_artefact(db, store, data)


# --------------------------------------------------------------------------------
# Dag execution
def execute(
    *outputs: Union[Operation, Artefact, ShellOutput],
    connection: Connection = None,
) -> None:
    """Trigger execution of a workflow to obtain a given output.

    Args:
        *outputs: Final artefacts or operations to be evaluated in the
            workflow. These objects and all of their dependencies will be
            executed by workers.
        connection: An explicit Redis connection. Not required if called
            within a `Fun()` context.
    """
    # get redis
    db, _ = get_connection(connection)

    # run dag
    for el in outputs:
        start_dag_execution(db, el.hash)


# --------------------------------------------------------------------------------
# Shell command
def shell(
    *args: str,
    inp: _INP_FILES = None,
    out: _OUT_FILES = None,
    env: Optional[dict[str, str]] = None,
    strict: bool = True,
    opt: Optional[Options] = None,
    connection: Connection = None,
) -> ShellOutput:
    """Add a shell command to the workflow.

    `shell()` puts a shell command in the workflow and returns a `types.ShellOutput`
    instance that provides a convenient wrapper to stdout, stderr and output
    files.

    Input and output files need to be explicitly given as arguments `inp` and
    `out`. Input and output files containing path separators (`/`) are assumed
    to belong to the corresponding directory tree structures, which will be
    automatically generated for input files.

    The `strict` flag determines how to interpret errors in input files. If
    `True` (the default), errors are propagated down: shell commands will not
    be executed if any input values currently hold `errors.Error`. Instead, all
    output values will also be replaced by `errors.Error`.

    When `strict=False`, input files with errors will simply (and silently) be
    excluded from the shell script.

    Shell commands are run in a temporary directory which conveys some measure
    of encapsulation, but it is quite weak, so the callee should make sure
    that commands only use relative paths etc. to ensure proper cleanup and
    function purity. This is done using python's `tempfile` module: the temporary
    directory can be set using the $TMPDIR environment variable.

    Environment variables can be passed to the executed command with the
    `env=` keyword. In contrast with `subprocess.Popen()`, the environment of
    worker processes will be updated with those values, *not* replaced by
    them. Environment variables are not hashed as part of the operation's id
    and thus changing them will not result in workflow re-execution.

    Args:
        *args: Lines of shell script to be evaluated.
        inp: Input files to pass to the shell comand. This should be a Mapping
            from filenames (str, path etc.) to values. Values can either be
            `types.Artefact` instances or of type `bytes`, in which case they
            will be automatically converted using `put()`.
        out: Filenames of output files that will be used to populate the return
            `types.ShellOutput` object. Note that any file not included in
            this list will be deleted when the shell command terminates.
        env: Environment variables to be set before calling the shell command.
        strict: If `False`, error handling will be deferred to the shell command
            by not populating input files of type `Error`.
        connection: An explicit Redis connection. Not required if called within a
            `Fun()` context.
        opt: An `types.Options` instance as returned by `options()`. Not
            required if called within a `Fun()` context.

    Returns:
        A `types.ShellOutput` object, populated with the generated
        `types.Artefact` instances.

    Raises:
        TypeError: when types of arguments are wrong.

    """
    opt = get_options(opt)
    db, store = get_connection(connection)

    # Parse args --------------------------------------------
    cmds: list[str] = []
    inputs: dict[str, Artefact] = {}

    for arg in args:
        if isinstance(arg, str):
            cmds += [arg]
        else:
            raise TypeError(f"argument {arg} not str.")

    # Parse input files -------------------------------------
    if inp is None:
        pass
    # multiple input files as a mapping
    elif isinstance(inp, Mapping):
        for key, val in inp.items():
            if isinstance(val, str):
                logger.warning(
                    f"{key} passed to shell as a string.\nif you don't want it to be"
                    + ' converted to json (and wrapped with "), \nyou NEED to pass it'
                    + " as bytes (by .encode()-ing it first)"
                )
            inputs[str(key)] = _artefact(db, store, val)
    else:
        raise TypeError(f"{inp} not a valid file input")

    if out is None:
        outputs = []
    else:
        outputs = [str(o) for o in out]

    inputs_types = dict([(k, v.kind) for k, v in inputs.items()])
    funsie = shell_funsie(cmds, inputs_types, outputs, env, strict=strict)
    operation = make_op(db, funsie, inputs, opt)
    return ShellOutput(db, operation)


# --------------------------------------------------------------------------------
# Data loading and saving
def put(
    value: T,
    *,
    connection: Connection = None,
) -> Artefact[T]:
    """Save data to Redis and return an Artefact.

    `put()` explicitly saves `value`, a bytes or string value, to the database
    and return an `types.Artefact` pointing to this value.

    The returned artefact's status is `types.ArtefactStatus.const` and its
    parent hash is `root`. This means that:

    - The arterfact is populated before any workflow operation is executed.
    - It has no dependencies
    - It is hashed according to content, not history.

    Thus, `put()` is used to set input values to workflows.

    Args:
        value: Data to be held in database. `str` data is encoded to `bytes`.
        connection (optional): An explicit Redis connection. Not required if
            called within a `Fun()` context.

    Returns:
        An `types.Artefact` instance with status `const`.
    """
    db, store = get_connection(connection)
    return _artefact(db, store, value)


def __log_error(where: hash_t, dat: Result[object]) -> None:
    if isinstance(dat, Error):
        logger.warning(f"data error at hash {shorten_hash(where)}")


# fmt:off
@overload
def take(where: Artefact[T], *, strict: Literal[True] = True, connection: Connection=None) -> T:  # noqa
    ...


@overload
def take(where: Artefact[T], *, strict: Literal[False] = False, connection: Connection=None) -> Result[T]:  # noqa
    ...
# fmt:on


def take(
    where: Artefact[T],
    *,
    strict: bool = True,
    connection: Connection = None,
) -> Union[T, Result[T]]:
    """Take data corresponding to a given artefact from Redis.

    `take()` returns the currently held value of pointed to by the
    `types.Artefact` instance `where` as `bytes`.

    If `strict=True` (the default) and `where` points to an `types.Error`
    value, this function will raise `errors.UnwrapError`. This is equivalent
    to running `unwrap()` on the return value.

    However if `strict=False`, the return value of `take()` is a
    `errors.Result[bytes]` variable, that is, either an instance of `bytes` or
    whatever `types.Error` is currently held by `where`.

    Finally, if `where` does not point to a valid redis-backed
    `types.Artefact` an `errors.Error` is returned of kind
    `errors.ErrorKind.Mismatch`.

    Args:
        where: `types.Artefact` pointer to data taken from the database.
        strict: If `False`, return a value of type `errors.Result[bytes]`.
        connection: An explicit Redis connection. Not required if called
            within a `Fun()` context.

    Returns:
        Either `bytes` or `errors.Result[bytes]` depending on strictness.

    Raises:
        errors.UnwrapError:
            if `where` contains an `errors.Error` and `strict=True`.

    """
    db, store = get_connection(connection)
    dat = get_data(db, store, where)
    __log_error(where.hash, dat)
    if strict:
        return unwrap(dat)
    else:
        return dat


def takeout(
    where: Artefact, filename: _AnyPath, *, connection: Connection = None
) -> None:  # noqa:DAR101,DAR201
    """`take()` an artefact and save it to `filename`.

    This is syntactic sugar around `take()`. This function is always strict.
    """
    db, store = get_connection(connection)
    dat = get_stream(db, store, where.hash)
    __log_error(where.hash, dat)
    dat = unwrap(dat)
    with open(filename, "wb") as f:
        shutil.copyfileobj(dat, f)


def wait_for(
    thing: Union[ShellOutput, Artefact, Operation],
    timeout: Optional[float] = None,
    *,
    connection: Connection = None,
) -> None:
    """Block execution until an artefact is generated or an operation is executed.

    Args:
        thing: `types.Artefact` or operation to wait on.
        timeout (optional): Number of seconds to wait for before raising an
            exception. If unspecified, timeout is taken to be infinite.
        connection (optional): An explicit Redis connection. Not required if
            called within a `Fun()` context.

    Raises:
        TimeoutError: if timeout is exceeded.
    """
    db, _ = get_connection(connection)
    if isinstance(thing, Artefact):

        def __stat() -> bool:
            return get_status(db, thing.hash, resolve_links=True) > 0

    else:
        if isinstance(thing, Operation):
            op = thing
        else:
            op = thing.op

        def __stat() -> bool:
            return is_it_cached(db, op)

    t0 = time.time()
    while True:
        t1 = time.time()

        if __stat():
            return

        if timeout is not None:
            if t1 - t0 > timeout:
                raise TimeoutError(
                    f"waited on {shorten_hash(thing.hash)} " + f"for {t1-t0} seconds."
                )

        # avoids hitting the DB way too often
        time.sleep(0.3)


def reset(
    thing: Union[ShellOutput, Operation, Artefact],
    *,
    recursive: bool = True,
    connection: Connection = None,
) -> None:
    """Reset data associated with an operation and its dependents.

    This function deletes data associated with an operation or the operation
    generating a given artefact without actually removing it from the
    workflow. This is useful if an operation failed due to circumstances
    outside of the control of `funsies`, such as a non-reproducible step or
    worker setup error. When the workflow is executed again, all the `reset()`
    steps will be re-computed.

    By default, `reset()` is applied recursively to all dependents of an
    operation.

    Args:
        thing: Operation to reset. If an `types.Artefact` is given, its parent
            operation is `reset()`.
        recursive: If False, only this operation is reset; its dependents are
            untouched. Note that this is dangerous, as it can make
            non-reproducible workflows.
        connection: An explicit Redis connection. Not required if called
            within a `Fun()` context.

    Raises:
        AttributeError:
            when an `types.Artefact` is reset that has status
            `types.ArtefactStatus.const`.
    """
    db, _ = get_connection(connection)
    if isinstance(thing, Artefact):
        h = thing.parent
        if h == "root":
            raise AttributeError("attempted to delete a const artefact.")
    else:
        h = thing.hash

    # Delete everything from the operation
    op = Operation.grab(db, h)
    for art in op.out.values():
        delete_artefact(db, art)

    if recursive:
        # and its dependencies
        for el in descendants(db, h):
            op = Operation.grab(db, el)
            for art in op.out.values():
                delete_artefact(db, art)



"""Serializers and deserializers for Encoding."""
from __future__ import annotations

# std
import json
import traceback
from typing import Any, Optional

# module
from ._constants import Encoding, hash_t
from .errors import Error, ErrorKind, Result


def decode(
    enc: Encoding,
    data: Result[bytes],
    carry_error: Optional[hash_t] = None,
) -> Result[Any]:
    """Deserialize data from bytes according to an Encoding."""
    if isinstance(data, Error):
        # forward errors
        return data

    if not isinstance(data, bytes):
        # duck typing classic
        return Error(  # type:ignore
            kind=ErrorKind.WrongType,
            details=f"attempted to decode value of type {type(data)} "
            + f"with decoder for {enc}",
            source=carry_error,
        )

    if enc == Encoding.blob:
        return data
    else:
        if enc == Encoding.json:
            try:
                return json.loads(data.decode())
            except Exception:
                tb_exc = traceback.format_exc()
                return Error(
                    kind=ErrorKind.JSONDecodingError, details=tb_exc, source=carry_error
                )
        else:
            return Error(
                kind=ErrorKind.UnknownEncodingError,
                details=f"encoding {enc} is not implemented.",
                source=carry_error,
            )


def encode(enc: Encoding, data: Result[object]) -> Result[bytes]:
    """Serialize data into bytes according to an Encoding."""
    if isinstance(data, Error):
        return data

    if enc == Encoding.blob:
        if not isinstance(data, bytes):
            # duck typing classic
            return Error(
                kind=ErrorKind.WrongType,
                details=f"attempted to encode value of type {type(data)} "
                + f"with encoder for {enc}",
            )
        else:
            return data

    elif enc == Encoding.json:
        try:
            return json.dumps(data).encode()
        except Exception:
            tb_exc = traceback.format_exc()
            return Error(kind=ErrorKind.JSONEncodingError, details=tb_exc)
    else:
        return Error(
            kind=ErrorKind.UnknownEncodingError,
            details=f"encoding {enc} is not implemented.",
        )


def kind(data: Any) -> Encoding:
    """Autodetect the correct encoder."""
    if isinstance(data, bytes):
        return Encoding.blob
    else:
        return Encoding.json



"""Run shell commands using funsies."""
from __future__ import annotations

# std
from io import BytesIO
import json
import os
import shutil
import subprocess
import tempfile
import time
from typing import Any, Mapping, Optional, Sequence

# external
from redis import Redis

# module
from ._constants import _Data, Encoding, hash_t
from ._funsies import Funsie, FunsieHow
from ._graph import Artefact, Operation
from ._logging import logger
from .errors import Error, Result

# Special namespaced "files"
SPECIAL = "__special__"
STDOUT = f"{SPECIAL}/stdout"
STDERR = f"{SPECIAL}/stderr"
RETURNCODE = f"{SPECIAL}/returncode"


def shell_funsie(
    cmds: Sequence[str],
    input_files: dict[str, Encoding],
    output_files: Sequence[str],
    env: Optional[dict[str, str]] = None,
    strict: bool = True,
) -> Funsie:
    """Wrap a shell command."""
    out = {}
    for fn in output_files:
        out[fn] = Encoding.blob
        # TODO: files that end with .json

    for k in range(len(cmds)):
        out[f"{STDOUT}{k}"] = Encoding.blob
        out[f"{STDERR}{k}"] = Encoding.blob
        out[f"{RETURNCODE}{k}"] = Encoding.json

    ierr = 1
    if strict:
        ierr = 0

    extra = dict(cmds=json.dumps(list(cmds)).encode(), env=json.dumps(env).encode())

    return Funsie(
        how=FunsieHow.shell,
        what=" && ".join(cmds),
        inp=input_files,
        out=out,
        error_tolerant=ierr,
        extra=extra,
    )


def run_shell_funsie(  # noqa:C901
    funsie: Funsie, input_values: Mapping[str, Result[BytesIO]]
) -> dict[str, Optional[_Data]]:
    """Execute a shell command."""
    logger.info("shell command")
    with tempfile.TemporaryDirectory() as dir:
        # Put in dir the input files
        for fn, val in input_values.items():
            if isinstance(val, Error):
                pass
            else:
                with open(os.path.join(dir, fn), "wb") as f:
                    shutil.copyfileobj(val, f)

        cmds = json.loads(funsie.extra["cmds"].decode())
        new_env = json.loads(funsie.extra["env"].decode())
        env: Optional[dict[str, str]] = None
        if new_env:
            env = os.environ.copy()
            env.update(new_env)

        out: dict[str, Optional[_Data]] = {}

        for k, c in enumerate(cmds):
            t1 = time.time()
            logger.info(f"{k+1}/{len(cmds)} $> {c}")
            proc = subprocess.run(c, cwd=dir, capture_output=True, shell=True, env=env)
            t2 = time.time()
            logger.info(f"done {k+1}/{len(cmds)} \t\tduration: {t2-t1:.2f}s")

            out[f"{STDOUT}{k}"] = proc.stdout
            out[f"{STDERR}{k}"] = proc.stderr
            out[f"{RETURNCODE}{k}"] = int(proc.returncode)
            if proc.returncode:
                logger.warning(f"nonzero returncode={proc.returncode}")

        # Output files
        for file in funsie.out:
            if SPECIAL in file:
                pass
            else:
                try:
                    with open(os.path.join(dir, file), "rb") as f:
                        out[file] = f.read()
                except FileNotFoundError:
                    logger.warning(f"missing expected output {file}")
                    out[file] = None
    return out


# --------------------------------------------------------------------------------
# Convenience class for shell funsie
class ShellOutput:
    """A convenience wrapper for a shell operation."""

    op: Operation
    hash: hash_t
    out: dict[str, Artefact[bytes]]
    inp: dict[str, Artefact[Any]]

    def __init__(self: "ShellOutput", store: Redis[bytes], op: Operation) -> None:
        """Generate a ShellOutput wrapper around a shell operation."""
        # import the constants
        # module
        from ._shell import RETURNCODE, SPECIAL, STDERR, STDOUT

        # stuff that is the same
        self.op = op
        self.hash = op.hash

        self.out = {}
        self.n = 0
        for key, val in op.out.items():
            if SPECIAL in key:
                if RETURNCODE in key:
                    self.n += 1  # count the number of commands
            else:
                self.out[key] = Artefact[bytes].grab(store, val)

        self.inp = {}
        for key, val in op.inp.items():
            self.inp[key] = Artefact[Any].grab(store, val)

        self.stdouts = []
        self.stderrs = []
        self.returncodes = []
        for i in range(self.n):
            self.stdouts += [Artefact[bytes].grab(store, op.out[f"{STDOUT}{i}"])]
            self.stderrs += [Artefact[bytes].grab(store, op.out[f"{STDERR}{i}"])]
            self.returncodes += [Artefact[int].grab(store, op.out[f"{RETURNCODE}{i}"])]

    def __check_len(self: "ShellOutput") -> None:
        if self.n > 1:
            raise AttributeError(
                "More than one shell command are included in this run."
            )

    @property
    def returncode(self: "ShellOutput") -> Artefact[int]:
        """Return code of a shell command."""
        self.__check_len()
        return self.returncodes[0]

    @property
    def stdout(self: "ShellOutput") -> Artefact[bytes]:
        """Stdout of a shell command."""
        self.__check_len()
        return self.stdouts[0]

    @property
    def stderr(self: "ShellOutput") -> Artefact[bytes]:
        """Stderr of a shell command."""
        self.__check_len()
        return self.stderrs[0]



"""graphviz related utilities."""
from __future__ import annotations

# std
import itertools
import os.path
from typing import Dict, Optional

# external
from redis import Redis

# module
from ._constants import ARTEFACTS, FUNSIES, hash_t, join
from ._dag import ancestors, get_nearest_operation
from ._graph import ArtefactStatus, Operation, resolve_link
from ._logging import logger
from ._short_hash import shorten_hash

__node_type = Dict[hash_t, Dict[str, Dict[str, hash_t]]]
__artefact_type = Dict[hash_t, ArtefactStatus]
__label_type = Dict[str, str]
__link_type = Dict[hash_t, hash_t]

# Watch out: nasty code ahead. This functionality is quite secondary to the
# program, so it's not written with the most care. Likely, the following will
# basically need a full rewrite eventually. If it breaks... maybe we just
# chuck it.


def __sanitize_command(lab: str) -> str:
    if ".<locals>" in lab:
        lab = lab.split(".<locals>")[0]
    return (
        lab.replace("<", r"\<")
        .replace(">", r"\>")
        .replace("|", r"\|")
        .replace("{", r"\{")
        .replace("}", r"\}")
        .replace('"', r"\"")
        .replace("&&", r"\n")
    )


def export(
    db: Redis[bytes], addresses: list[hash_t]
) -> tuple[__node_type, __artefact_type, __label_type, __link_type]:
    """Output a DAG in dot format for graphviz."""
    nodes: __node_type = {}
    labels: dict[str, str] = {}
    artefacts: __artefact_type = {}

    funsies = {}
    art_hashes = set()

    for address in addresses:
        node = get_nearest_operation(db, address)
        if node is None:
            continue

        curr_nodes = ancestors(db, node.hash, include_subdags=True)
        curr_nodes.add(node.hash)
        logger.info(f"graph contains {len(curr_nodes)} nodes")

        for k, h in enumerate(curr_nodes):
            if k % 100 == 0 and k > 0:
                logger.info(f"done: {k} / {len(curr_nodes)}")
            # all the operations
            nodes[h] = {}
            obj = Operation.grab(db, h)

            # get funsies data, cache it too
            if obj.funsie not in funsies:
                funsies[obj.funsie] = db.hgetall(join(FUNSIES, obj.funsie))
            funsie = funsies[obj.funsie]

            labels[h] = __sanitize_command(funsie[b"what"].decode())
            if funsie[b"error_tolerant"] == b"1":
                labels[h] += r"\n(tolerates err)"

            nodes[h]["inputs"] = obj.inp
            nodes[h]["outputs"] = obj.out

            for h in itertools.chain(obj.inp.values(), obj.out.values()):
                # cache status too
                art_hashes.add(h)

    logger.info(f"gathering status for {len(art_hashes)} artefacts")
    artefact_hashes = list(art_hashes)
    redis_keys = [join(ARTEFACTS, address, "status") for address in artefact_hashes]
    statuses: list[Optional[bytes]] = db.mget(redis_keys)
    for h, s in zip(artefact_hashes, statuses):
        if s is None:
            artefacts[h] = ArtefactStatus.not_found
        else:
            artefacts[h] = ArtefactStatus(int(s.decode()))

    # Get links
    links: dict[hash_t, hash_t] = {}
    for h, v in artefacts.items():
        if v == ArtefactStatus.linked:
            links[h] = resolve_link(db, h)
    logger.info(f"found {len(links)} links")

    return nodes, artefacts, labels, links


__style_line_link = "color=black,penwidth=8.0"


def __style_line(i: ArtefactStatus) -> str:
    if i == 1:
        return "color=2,penwidth=3.0"
    elif i <= 0:
        return "color=8,penwidth=3.0"
    elif i == 2:
        return "color=black,penwidth=1.0"
    elif i == 3:
        return "color=6,penwidth=3.0"
    else:
        return "color=black"


def __style_node(i: ArtefactStatus) -> str:
    if i == 1:
        return "style=filled,color=1"
    if i == 0:
        return "style=filled,color=7"
    if i == -2:
        return "style=filled,color=7,fontcolor=6"
    if i == 2:
        return "color=white"
    elif i == 3:
        return "style=filled,color=5"
    elif i == 4:
        return "color=white"
    else:
        return ""


def __sanitize_fn(n: str) -> str:
    return os.path.basename(n)


def format_dot(  # noqa:C901
    nodes: __node_type,
    artefacts: __artefact_type,
    labels: __label_type,
    links: dict[hash_t, hash_t],
    targets: list[hash_t],
    show_inputs: bool = True,
) -> str:
    """Format output of export() for graphviz dot."""
    header = (
        "digraph G {\nrankdir=LR;\n"
        + "node [fontsize=24, colorscheme=paired8];\n"
        + "edge [fontsize=24, colorscheme=paired8];\n"
    )

    keep: dict[hash_t, list[hash_t]] = {}
    finals: dict[hash_t, hash_t] = {}
    initials: dict[hash_t, list[hash_t]] = {}
    # Setup which artefacts to show, which are inputs and which are outputs.
    for n in nodes:
        for v in nodes[n]["inputs"].values():
            # todo make conditional keep setable
            if artefacts[v] == 2:
                initials[v] = initials.get(v, []) + [n]
                if show_inputs:
                    keep[v] = keep.get(v, []) + [n]
            else:
                keep[v] = keep.get(v, []) + [n]

        # if targets are artefacts, then we should always keep them
        for t in targets:
            if t in nodes[n]["outputs"].values():
                keep[t] = []
                finals[t] = n

        # if targets are nodes, then we should always keep all the artefacts.
        if n in targets:
            for v in nodes[n]["outputs"].values():
                keep[v] = keep.get(v, []) + [n]
                finals[v] = n

    # keep all the links too
    for k, v in links.items():
        keep[v] = keep.get(v, []) + [k]

    # write operation nodes
    nstring = ""
    for n in nodes:
        inps = []
        for kk, v in nodes[n]["inputs"].items():
            if v in keep:
                inps += [f"<A{v}>{__sanitize_fn(kk)}"]
        inps = "|".join(inps)

        outs = []
        for kk, v in nodes[n]["outputs"].items():
            if v in keep:
                outs += [f"<A{v}>{__sanitize_fn(kk)}"]
        outs = "|".join(outs)

        nstring += (
            f"N{n} ["
            + "shape=record,width=.1,height=.1,"
            + 'label="'
            + "{{"
            + f"{inps}"
            + "}|"
            + f" {shorten_hash(n)} \\n{labels[n]}"
            + "|{"
            + f"{outs}"
            + "}}"
            + '"];\n'
        )

    # write artefact nodes
    for k in set(keep):
        nstring += (
            f"A{k}"
            + f'[shape=box,label="{shorten_hash(k)}"'
            + f",{__style_node(artefacts[k])}];\n"
        )

    # Make connections
    connect = ""
    for n in nodes:
        for _, v in nodes[n]["outputs"].items():
            if v in keep:
                connect += f"N{n}:A{v} -> A{v} [{__style_line(artefacts[v])}];\n"

        for _, v in nodes[n]["inputs"].items():
            if v in keep:
                connect += f"A{v} -> N{n}:A{v} [{__style_line(artefacts[v])}];\n"

    ranks = ""
    for k, v in links.items():
        connect += f"A{v} -> A{k} [{__style_line_link}];\n"
        ranks += "{" + f"rank = same; A{v}; A{k};" + "}\n"

    if show_inputs:
        init = []
        for a in initials:
            init += [f"A{a}"]
        if len(init):
            ranks += "{rank = same;" + ";".join(init) + ";}\n"

    # rank_last = []
    # for t in finals.keys():
    #     rank_last += [f"A{t}"]
    # ranks += "{rank = same;" + ";".join(rank_last) + ";}\n"

    footer = "\n}"
    return header + nstring + connect + ranks + footer



"""Main data structures."""
from __future__ import annotations

# std
from dataclasses import dataclass, field
from enum import IntEnum
import hashlib
from io import BytesIO
from typing import cast, Mapping, Type

# external
from redis import Redis
from redis.client import Pipeline

# module
from . import _serdes
from ._constants import Encoding, FUNSIES, hash_t, join
from ._short_hash import hash_save
from .errors import Error, Result


def _artefacts(inp: dict[bytes, bytes]) -> dict[str, Encoding]:
    out = {}
    for key, val in inp.items():
        out[key.decode()] = Encoding(val.decode())
    return out


# --------------------------------------------------------------------------------
class FunsieHow(IntEnum):
    """Kinds of funsies.

    This enum contains the various kinds of funsies: python code, shell code,
    etc.
    """

    python = 0
    shell = 1
    subdag = 2


@dataclass
class Funsie:
    """A funsie is a wrapped command that can be backed up to the KV store.

    A Funsie has a `how`, a `what` (a string that identifies the funsie, such
    as a function name or shell commands) and input and output artefact names.
    All of these are used to generate the hash of the Funsie instance.

    Funsies also have an `extra` field that include auxiliary data that is not
    to be used in hashing the funsie, but is useful for executing it, such as
    a cloudpickled python function.

    """

    how: FunsieHow
    what: str
    inp: dict[str, Encoding]
    out: dict[str, Encoding]
    extra: dict[str, bytes]
    error_tolerant: int = 0
    hash: hash_t = field(init=False)

    def decode(
        self: Funsie, input_data: Mapping[str, Result[BytesIO]]
    ) -> dict[str, Result[object]]:
        """Decode input data according to `inp`."""
        out = {}
        for key, enc in self.inp.items():
            element = input_data[key]
            if isinstance(element, Error):
                out[key] = cast(Result, element)
            else:
                out[key] = _serdes.decode(enc, element.read())

            if self.error_tolerant == 0 and isinstance(out[key], Error):
                raise RuntimeError(
                    f"Decoding of input data {key} failed:\n{out[key].details}"
                )
        return out

    def put(self: Funsie, db: Redis[bytes]) -> None:
        """Save a Funsie to Redis."""
        db.hset(
            join(FUNSIES, self.hash),
            mapping={
                "hash": self.hash,
                "how": int(self.how),
                "what": self.what,
                "error_tolerant": self.error_tolerant,
            },
        )
        if self.inp:
            db.hset(
                join(FUNSIES, self.hash, "inp"),
                mapping=dict([(k, v.value) for k, v in self.inp.items()]),
            )
        if self.out:
            db.hset(
                join(FUNSIES, self.hash, "out"),
                mapping=dict([(k, v.value) for k, v in self.out.items()]),
            )
        if self.extra:
            db.hset(
                join(FUNSIES, self.hash, "extra"), mapping=self.extra  # type:ignore
            )

        # Save the hash in the quickhash db
        hash_save(db, self.hash)

    @classmethod
    def grab(cls: Type[Funsie], db: Redis[bytes], hash: hash_t) -> "Funsie":
        """Grab a Funsie from the Redis store."""
        pipe: Pipeline = db.pipeline(transaction=False)
        pipe.exists(join(FUNSIES, hash))
        pipe.hgetall(join(FUNSIES, hash))
        pipe.hgetall(join(FUNSIES, hash, "inp"))
        pipe.hgetall(join(FUNSIES, hash, "out"))
        pipe.hgetall(join(FUNSIES, hash, "extra"))
        exists, metadata, inp, out, extra = pipe.execute()

        if not exists:
            raise RuntimeError(f"No funsie at {hash}")

        return Funsie(
            how=FunsieHow(int(metadata[b"how"].decode())),
            what=metadata[b"what"].decode(),
            error_tolerant=int(metadata[b"error_tolerant"].decode()),
            inp=_artefacts(inp),
            out=_artefacts(out),
            extra=dict([(k.decode(), v) for k, v in extra.items()]),
        )

    def __str__(self: Funsie) -> str:
        """Get the string representation of a funsie."""
        # ==============================================================
        #     ALERT: DO NOT TOUCH THIS CODE WITHOUT CAREFUL THOUGHT
        # --------------------------------------------------------------
        # When hashes change, previous databases become deprecated. This
        # (will) require a change in version number!
        out = f"how={self.how}\n" + f"what={self.what}\n"
        for key in sorted(self.inp.keys()):
            out += f"input:{key} -> {self.inp[key].value}\n"
        for key in sorted(self.out.keys()):
            out += f"output:{key} -> {self.out[key].value}\n"
        out += f"error tolerant:{self.error_tolerant}\n"
        # ==============================================================
        return out

    def __post_init__(self: Funsie) -> None:
        """Calculate the hash."""
        # ==============================================================
        #     ALERT: DO NOT TOUCH THIS CODE WITHOUT CAREFUL THOUGHT
        # --------------------------------------------------------------
        # When hashes change, previous databases become deprecated. This
        # (will) require a change in version number!
        m = hashlib.sha1()
        # header
        m.update(b"funsie")
        # funsie
        m.update(str(self).encode())
        self.hash = hash_t(m.hexdigest())
        # ==============================================================



"""Parametrize DAGs and re-run them dynamically."""
from __future__ import annotations

# std
from dataclasses import dataclass
import hashlib
from typing import cast, Optional, Set, Type

# external
from redis import Redis
from redis.client import Pipeline

# module
from ._constants import ARTEFACTS, hash_t, join, OPERATIONS, PARAMETRIC
from ._dag import ancestors, descendants
from ._funsies import Funsie
from ._graph import Artefact, make_op, Operation
from ._logging import logger


def _parametrize_subgraph(
    db: Redis[bytes], inputs: dict[str, Artefact], outputs: dict[str, Artefact]
) -> set[hash_t]:
    """Return the subgraph of operators that connects given inputs and outputs."""
    # First, we grab all the ancestors for all the outputs.
    output_ops = set()
    for o in outputs.values():
        if o.parent == "root":
            raise RuntimeError("output of parametrization has no dependencies.")
        else:
            output_ops.add(o.parent)
    out_ancestors = ancestors(db, *output_ops).union(output_ops)

    # then we grab all the descendants for all the inputs.
    input_ops = set()
    for name, inp in inputs.items():
        inp_dependent_ops = []
        for dependent in db.smembers(join(ARTEFACTS, inp.hash, "dependents")):
            d: str = dependent.decode()
            if d in out_ancestors:
                inp_dependent_ops += [d]

        if len(inp_dependent_ops) == 0:
            logger.error(f"parametrized input {name} does not change any outputs!")

        input_ops.update(inp_dependent_ops)
    input_ops = cast(Set[hash_t], input_ops)
    in_descendants = descendants(db, *input_ops).union(input_ops)

    # The intersection between these two sets forms the set of operators that
    # we need in our parametrization.
    return out_ancestors.intersection(in_descendants)


def _subgraph_edges(db: Redis[bytes], nodes: set[hash_t]) -> dict[hash_t, set[hash_t]]:
    """Compute edges on a connected subgraph."""
    # get edges
    edges: dict[hash_t, set[hash_t]] = {}
    for n in nodes:
        for value in db.smembers(join(OPERATIONS, n, "parents")):
            parent: hash_t = hash_t(value.decode())
            if parent in nodes:
                edges[parent] = edges.get(parent, set()).union([n])
    return edges


def _subgraph_toposort(
    nodes: set[hash_t], edges: dict[hash_t, set[hash_t]]
) -> list[hash_t]:
    """Sort a connected subgraph topologically."""
    # from https://en.wikipedia.org/wiki/Topological_sorting#Kahn's_algorithm

    # remove all the nodes that have dependencies to get the initial set
    S = nodes.copy()
    for out in edges.values():
        for k in out:
            S.discard(k)

    output = []
    while len(S):
        node_n = S.pop()
        output.append(node_n)
        edge_set = edges.pop(node_n, set())

        # keep only the nodes with no dependencies
        for out in edges.values():
            for k in out:
                edge_set.discard(k)

        S = S.union(edge_set)
    return output


def _hash_parametric(
    db: Redis[bytes],
    sorted_nodes: list[hash_t],
    inputs: dict[str, Artefact],
    outputs: dict[str, Artefact],
) -> hash_t:
    """Get hash for a Parametric."""
    # This is done basically the exact same way as evaluating a parametric,
    # but hashing instead of evaluating.

    translation: dict[hash_t, hash_t] = {}  # old hash -> new artefact
    # bootstart using the new input values
    for name in inputs.keys():
        translation[inputs[name].hash] = hash_t(name)

    # these are sorted so this is fairly straightforward
    for node in sorted_nodes:
        op = Operation.grab(db, node)
        new_inp = {}
        for k, v in op.inp.items():
            if v not in translation:
                translation[v] = v
                # TODO: we could do artefact freezing here relatively easily.
            new_inp[k] = translation[v]

        # ==============================================================
        #     ALERT: DO NOT TOUCH THIS CODE WITHOUT CAREFUL THOUGHT
        # --------------------------------------------------------------
        # When hashes change, previous databases become deprecated. This
        # (will) require a change in version number!
        m = hashlib.sha1()
        m.update(op.funsie.encode())
        for key, val in sorted(new_inp.items(), key=lambda x: x[0]):
            m.update(f"file={key}, hash={val}".encode())
        ophash = hash_t(m.hexdigest())

        for key, val in op.out.items():
            translation[val] = hash_t(ophash + ":" + key)

    # STILL NO TOUCHING
    m = hashlib.sha1()
    m.update("parametric".encode())
    for key, art in sorted(outputs.items(), key=lambda x: x[0]):
        m.update(f"output:{key}, hash:{translation[art.hash]}".encode())

    return hash_t(m.hexdigest())


def _do_parametrize(
    db: Redis[bytes],
    sorted_nodes: list[hash_t],
    inputs: dict[str, hash_t],
    outputs: dict[str, hash_t],
    new_inputs: dict[str, Artefact],
) -> dict[str, Artefact]:
    """Take a subgraph and re-build it with input values changed for new_inputs."""
    translation = {}  # old hash -> new artefact
    # bootstart using the new input values
    for name, artefact in new_inputs.items():
        translation[inputs[name]] = artefact

    # these are sorted so this is fairly straightforward
    for node in sorted_nodes:
        op = Operation.grab(db, node)
        new_inp = {}
        for k, v in op.inp.items():
            if v not in translation:
                # update to minimize the amount of artefact grabbing
                translation[v] = Artefact.grab(db, v)
                # TODO: we could do artefact freezing here relatively easily.

            new_inp[k] = translation[v]

        funsie = Funsie.grab(db, op.funsie)
        assert op.options is not None  # not needed as it is never None
        new_op = make_op(db, funsie, new_inp, op.options)
        for k, v in new_op.out.items():
            # update translation map with new ops artefacts
            translation[op.out[k]] = Artefact.grab(db, v)

    return_value = {}
    for k, h in outputs.items():
        return_value[k] = translation[h]
    return return_value


@dataclass(frozen=True)
class Parametric:
    """Parametrized DAG."""

    name: str
    hash: hash_t
    ops: list[hash_t]
    inp: dict[str, hash_t]
    out: dict[str, hash_t]

    def evaluate(
        self: Parametric, db: Redis[bytes], new_inp: dict[str, Artefact]
    ) -> dict[str, Artefact]:
        """Generate DAG from a Parametric."""
        return _do_parametrize(db, self.ops, self.inp, self.out, new_inp)

    def put(self: Parametric, db: Redis[bytes]) -> None:
        """Save a Parametric DAG to Redis."""
        db.hset(join(PARAMETRIC, self.hash, "inp"), mapping=self.inp)  # type:ignore
        db.hset(join(PARAMETRIC, self.hash, "out"), mapping=self.out)  # type:ignore
        db.rpush(join(PARAMETRIC, self.hash), *self.ops)
        db.set(join(PARAMETRIC, self.hash, "name"), self.name.encode())

        # set the name for reverse lookups
        db.hset(
            join(PARAMETRIC, hash_t("names")), self.name.encode(), self.hash.encode()
        )

    @classmethod
    def grab(cls: Type[Parametric], db: Redis[bytes], hash: hash_t) -> Parametric:
        """Grab a Parametric DAG from the Redis store."""
        pipe: Pipeline = db.pipeline(transaction=False)
        pipe.exists(join(PARAMETRIC, hash))
        pipe.lrange(join(PARAMETRIC, hash), 0, -1)
        pipe.hgetall(join(PARAMETRIC, hash, "inp"))
        pipe.hgetall(join(PARAMETRIC, hash, "out"))
        pipe.get(join(PARAMETRIC, hash, "name"))
        exists, lrange, inp, out, name = pipe.execute()

        if not exists:
            raise RuntimeError(f"No parametric DAG at {hash}")

        name = name.decode()
        ops = [hash_t(el.decode()) for el in lrange]

        return Parametric(
            name=name,
            hash=hash,
            ops=ops,
            inp=dict([(k.decode(), hash_t(v.decode())) for k, v in inp.items()]),
            out=dict([(k.decode(), hash_t(v.decode())) for k, v in out.items()]),
        )

    @classmethod
    def resolve_name(
        cls: Type[Parametric], db: Redis[bytes], name: str
    ) -> Optional[hash_t]:
        """Get the parametric hash corresponding to a given name."""
        h = db.hget(join(PARAMETRIC, hash_t("names")), name)
        if h is None:
            return None
        else:
            return hash_t(h.decode())


def make_parametric(
    db: Redis[bytes],
    name: str,
    inp: dict[str, Artefact],
    out: dict[str, Artefact],
) -> Parametric:
    """Define a parametric DAG."""
    # Get the ops
    ops = _parametrize_subgraph(db, inp, out)

    # Sort topologically
    edges = _subgraph_edges(db, ops)
    sorted_ops = _subgraph_toposort(ops, edges)

    param = Parametric(
        name=name,
        hash=_hash_parametric(db, sorted_ops, inp, out),
        ops=sorted_ops,
        inp=dict([(k, v.hash) for k, v in inp.items()]),
        out=dict([(k, v.hash) for k, v in out.items()]),
    )

    # Save in db
    pipe: Pipeline = db.pipeline(transaction=False)
    param.put(pipe)
    pipe.execute()
    return param



"""Generic getter of stuff."""
from __future__ import annotations

# std
from typing import Union

# module
from . import _constants as c
from ._context import Connection, get_connection
from ._funsies import Funsie
from ._graph import Artefact, Operation
from ._logging import logger
from ._short_hash import hash_load


def get(
    target: str,
    connection: Connection = None,
) -> list[Union[Artefact, Funsie, Operation]]:
    """Get object or objects that correspond to a given hash value.

    `get()` returns a list of objects (`Artefact`, `Operation` and `Funsie`
    instances) currently on the active Redis connection that have a hash
    address starting with `target`. This function allows programatically
    retrieving hashes like the ```funsies cat``` command does.

    Args:
        target: A hash or truncated hash value.
        connection (optional): An explicit Redis connection. Not required if
            called within a `Fun()` context.

    Returns:
        A list of objects with ids that start with `target`. Empty if no such
        objects exist.
    """
    db, store = get_connection(connection)
    hashes = hash_load(db, target)
    out: list[Union[Artefact, Funsie, Operation]] = []
    for h in hashes:
        if db.exists(c.join(c.ARTEFACTS, h)):
            logger.debug(f"{h} is Artefact")
            out += [Artefact.grab(db, h)]

        elif db.exists(c.join(c.FUNSIES, h)):
            logger.debug(f"{h} is Funsie")
            out += [Funsie.grab(db, h)]

        elif db.exists(c.join(c.OPERATIONS, h)):
            logger.debug(f"{h} is Operation")
            out += [Operation.grab(db, h)]

        else:
            logger.debug(f"{h} does not exist")
    return out



"""Logging and log levels."""
# std
import sys

# external
import loguru

# logger = logging.getLogger("funsies")
logger = loguru.logger
logger.remove()


# Standard logger for errors etc
def set_level(level: str) -> None:
    """Set logging level."""
    log_format = (
        "<green>{time:YYYYMMDD} {time:HH:mm:ss.S}</green>|"
        + " <level>{level: <8}</level> |"
        + " <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan>"
        + " <level>{message}</level>"
    )
    logger.remove()
    logger.add(
        sys.stderr,
        format=log_format,
        level=level,
        filter=lambda record: "op" not in record["extra"],
    )


# Default level
set_level("INFO")

# Error logger for operations
worker_format = " op:<green>{extra[op]}</green> |" + " <level>{message}</level>"
logger.add(
    sys.stderr,
    format=worker_format,
    filter=lambda record: "op" in record["extra"],
    level="INFO",
    backtrace=False,
)



"""Funsies is a lightweight workflow engine 🔧.

.. include:: documentation.md
"""
# module
from . import debug, dynamic, parametric, types, utils
from ._context import Fun, ManagedFun, options
from ._getter import get
from .errors import unwrap
from .fp import morph, py, reduce
from .template import template
from .ui import execute, put, reset, shell, take, takeout, wait_for

__all__ = [
    # shell
    "shell",
    # fp
    "py",
    "reduce",
    "morph",
    # template
    "template",
    # artefact manipulation
    "take",
    "takeout",
    "put",
    "execute",
    "wait_for",
    "reset",
    "get",
    # contexts
    "Fun",
    "ManagedFun",
    "options",
    # Error handling and types
    "unwrap",
    "types",
    "debug",
    # utility
    "utils",
    "dynamic",
    "parametric",
]


# Version information
# We grab it from setup.py so that we don't have to bump versions in multiple
# places.
try:
    # std
    from importlib import metadata

    __version__ = metadata.version("funsies")
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    # external
    import importlib_metadata

    __version__ = importlib_metadata.version("funsies")



"""Names of stuff in the key value store."""
from __future__ import annotations

# std
from enum import Enum
from os import PathLike
from typing import NewType, Union

JsonData = Union[str, int, float, bool, None, dict, list]
"""Can be converted to JSON."""

_Data = Union[bytes, JsonData]
"""All output data types."""

# Some types
_AnyPath = Union[str, PathLike]
hash_t = NewType("hash_t", str)


class Encoding(str, Enum):
    """Types for data objects.

    Funsies does not support a full-blown type system. For this, we defer to
    json's encoding of various data structures.
    """

    json = "json"
    blob = "blob"


# Utility functions
def join(prefix: str, address: hash_t, *suffix: str) -> str:
    """Make a redis identifier."""
    return ":".join([prefix, str(address)] + list(suffix))


# Some locations
ARTEFACTS = "funsies.artefacts"
FUNSIES = "funsies.funsies"
OPERATIONS = "funsies.ops"
PARAMETRIC = "funsies.parametric"
HASH_INDEX = "funsies.hash_index"

# DAGs
DAG_OPERATIONS = "funsies.dags.operations"
DAG_STATUS = "funsies.dags.status.done"
DAG_INDEX = "funsies.dags.index"



"""Configuration dictionaries for jobs."""
from __future__ import annotations

# std
from dataclasses import asdict, dataclass
import json
import os
import sys
from typing import Any, Mapping, Optional, Type

# external
# redis
from redis import Redis

# module
from ._logging import logger
from ._storage import DiskStorage, RedisStorage, StorageEngine

# Constants
INFINITE = -1
ONE_DAY = 86400
ONE_MINUTE = 60


def _redis_connection(url: str, try_fail: bool = True) -> Redis[bytes]:
    """Open a new redis connection."""
    hn = _extract_hostname(url)
    logger.info(f"connecting to {hn}")
    db = Redis.from_url(url, decode_responses=False)
    try:
        db.ping()
    except Exception as e:
        if try_fail:
            raise e
        else:
            logger.error("could not connect to server! exiting")
            logger.error(str(e))
            sys.exit(-1)
    logger.debug("connection sucessful")
    logger.success(f"connected to {hn}")
    return db


# --------------------------------------------------------------------------
# Redis configuration
# --------------------------------------------------------------------------
def _get_redis_url(url: Optional[str] = None) -> str:
    """Get the default funsies redis URL."""
    if url is not None:
        return url
    else:
        try:
            default = os.environ["FUNSIES_JOBS"]
        except KeyError:
            try:
                default = os.environ["FUNSIES_URL"]
            except KeyError:
                default = "redis://localhost:6379"

    return default


def _get_storage_url(url: Optional[str] = None) -> Optional[str]:
    """Get the default funsies storage URL."""
    if url is not None:
        return url
    else:
        try:
            default: Optional[str] = os.environ["FUNSIES_DATA"]
        except KeyError:
            default = None
        return default


def _extract_hostname(url: str) -> str:
    """Get the hostname part of the url."""
    if "@" in url:
        hn = url.split("@")[-1]
    else:
        hn = url.split("//")[-1]
    return hn


class Server:
    """Runtime options for the funsies redis server."""

    # Connection settings
    jobs_url: str
    data_url: str

    def __init__(
        self: Server,
        jobs_url: Optional[str] = None,
        data_url: Optional[str] = None,
    ) -> None:
        """Create a new funsies server configuration."""
        # defaults to the env variable FUNSIES_URL
        self.jobs_url = _get_redis_url(jobs_url)
        data_url = _get_storage_url(data_url)
        if data_url is None:
            self.data_url = self.jobs_url
        else:
            self.data_url = data_url

    def new_connection(
        self: Server, try_fail: bool = True
    ) -> tuple[Redis[bytes], StorageEngine]:
        """Setup job server and data connections."""
        rdb = _redis_connection(self.jobs_url, try_fail)
        if self.data_url == self.jobs_url:
            store: StorageEngine = RedisStorage(rdb)

        elif self.data_url[:5] == "redis":
            logger.info(f"saving artefacts to {self.data_url}")
            rstore = _redis_connection(self.data_url, try_fail)
            store = RedisStorage(rstore)

        elif self.data_url[:7] == "file://":
            store = DiskStorage(self.data_url[7:])

        else:
            raise NotImplementedError(
                f"No storage protocol corresponding to {self.data_url}\n"
                + "did you forget to start the url with redis:// or file://?"
            )

        return rdb, store


class MockServer(Server):
    """Mock redis server using FakeRedis for testing."""

    # Connection settings
    redis_url: str
    _instance: Redis[bytes]

    def __init__(self: MockServer, redis_url: Optional[str] = None) -> None:
        """Create a new funsies server configuration."""
        if redis_url is None:
            self.redis_url = "fakeredis://mock_server"
        else:
            self.redis_url = redis_url

        # make the connection
        # external
        from fakeredis import FakeStrictRedis as Redis

        self._instance = Redis()

    def new_connection(
        self: MockServer, try_fail: bool = True
    ) -> tuple[Redis[bytes], StorageEngine]:
        """Create a new redis connection."""
        return self._instance, RedisStorage(self._instance)


# --------------------------------------------------------------------------
# Job configuration
# --------------------------------------------------------------------------
@dataclass
class Options:
    """Runtime options for an Operation.

    This is a class that basically contains all the random options that may
    need to be set when building a workflow, such as timeouts, heterogeneous
    compute etc. It should generally be instantiated using the
    `funsies.options()` function.

    """

    timeout: int = INFINITE
    """Max execution time for this operation, in seconds or -1 for an operation
    that never timeouts. Defaults to -1."""

    queue: str = "default"
    """Defines which queue this operation should be executed by. For example,
    if a complex workflow requires GPUs for certain jobs, those jobs would be
    setup with option `queue="gpu"` and workers on the nodes with available
    GPUs would be instantiated with `funsies worker gpu`. Then, only worker
    processes in the GPU queue would execute the GPU jobs."""

    distributed: bool = True
    """If False, jobs are executed by the local enqueuing process. Used to
    test workflows without having to start workers."""

    reset: bool = False
    """If `True`, this operation is `funsies.reset()` when generated."""

    evaluate: bool = True
    """If False, calling `funsies.execute()` on this job or its dependencies will fail.
    Can be used to ensure a specific branch is never executed."""

    ttl: int = ONE_DAY
    """Time to live (ttl) in queue for the operation. Defaults to 24h. Equivalent
    to the [rq keyword with the same name](https://python-rq.org/docs/). """

    result_ttl: int = ONE_MINUTE
    """Time to live (ttl) in queue for the rq result objects. Defaults to one
    minute. Equivalent to the [rq keyword with the same
    name](https://python-rq.org/docs/). (Note that this has nothing to do with
    the actual data results.) """

    failure_ttl: int = ONE_DAY
    """Time to live (ttl) in queue for the rq result objects of failing jobs.
    Defaults to one day. Equivalent to the [rq keyword with the same
    name](https://python-rq.org/docs/). (Note that this has nothing to do with
    the actual data results.) """

    # TODO: make meaningfully adjustable
    serializer: str = "rq.serializers.JSONSerializer"

    @property
    def job_args(self: "Options") -> Mapping[str, Any]:
        """Return a dictionary of arguments for rq.enqueue's job_args."""
        return dict(
            timeout=self.timeout,
            ttl=self.ttl,
            result_ttl=self.result_ttl,
            failure_ttl=self.failure_ttl,
        )

    @property
    def task_args(self: "Options") -> Mapping[str, Any]:
        """Return a dictionary of arguments for dag.task()."""
        return dict(
            evaluate=self.evaluate,
        )

    @property
    def queue_args(self: "Options") -> Mapping[str, Any]:
        """Return a dictionary of arguments for rq.Queue."""
        return dict(is_async=self.distributed, serializer=self.serializer)

    def pack(self: "Options") -> str:
        """Pack an Options instance to a bytestring."""
        return json.dumps(asdict(self))

    @classmethod
    def unpack(cls: Type["Options"], data: str) -> "Options":
        """Unpack an Options instance from a byte string."""
        return Options(**json.loads(data))



"""Refer to objects by smaller hashes."""
from __future__ import annotations

# std
from typing import Union

# external
from redis import Redis

# module
from ._constants import HASH_INDEX, hash_t
from ._logging import logger

# Constants
SHORT = 6  # Length of short hashes


def shorten_hash(h: Union[hash_t, str]) -> str:
    """Shorten a hash."""
    return h[:SHORT]


def hash_save(db: Redis[bytes], hash: hash_t) -> None:
    """Save the shortened version of this hash for convenient retrieval."""
    db.zadd(HASH_INDEX, {str(hash): "0"})


def hash_load(db: Redis[bytes], short_hash: str) -> list[hash_t]:
    """Save the shortened version of this hash for convenient retrieval."""
    if len(short_hash) > 40:
        raise AttributeError(f"hash {short_hash} has length {len(short_hash)} > 40")

    data = db.zrangebylex(HASH_INDEX, f"[{short_hash}", "+")
    out = []
    for key in data:
        k = key.decode()
        if k[: len(short_hash)] == short_hash:
            out += [hash_t(k)]

    if len(out) == 0:
        logger.error(f"{short_hash} not found")
    elif len(out) > 1:
        logger.warning(f"{len(data)} possible values for {short_hash}")
    return out



"""Dynamic DAG generation."""
from __future__ import annotations

# std
from typing import Any, Callable, Optional, Sequence, TypeVar

# module
from ._constants import _Data, Encoding
from ._context import Connection, get_connection, get_options
from ._graph import Artefact, constant_artefact, make_op
from ._subdag import subdag_funsie
from .config import Options
from .ui import _Target, put

# --------------------------------------------------------------------------------
Tin = TypeVar("Tin", bound=_Data)
T1 = TypeVar("T1", bound=_Data)
T2 = TypeVar("T2", bound=_Data)
T3 = TypeVar("T3", bound=_Data)


def sac(
    split_fun: Callable[..., Sequence[T1]],
    apply_fun: Callable[[Artefact[T1]], Artefact[T2]],
    combine_fun: Callable[[Sequence[Artefact[T2]]], Artefact[T3]],
    *inp: _Target,
    out: Encoding,
    name: Optional[str] = None,
    strict: bool = True,
    opt: Optional[Options] = None,
    connection: Connection = None,
) -> Artefact[T3]:
    """Perform a split/apply/combine dynamic DAG."""
    opt = get_options(opt)
    db, store = get_connection(connection)

    inputs: dict[str, Artefact] = {}
    # Parse input  -------------------------------------
    inputs = {}
    arg_names = []
    for k, arg in enumerate(inp):
        arg_names += [f"in{k}"]
        if isinstance(arg, Artefact):
            inputs[arg_names[-1]] = arg
        else:
            inputs[arg_names[-1]] = put(arg, connection=db)
    inp_types = dict([(k, val.kind) for k, val in inputs.items()])

    if name is not None:
        fun_name = name
    else:
        fun_name = (
            f"SAC|{split_fun.__qualname__}|{apply_fun.__qualname__}"
            + f"|{combine_fun.__qualname__}"
        )

    def __sac(inpd: dict[str, Any]) -> dict[str, Artefact[T3]]:
        """Perform the map reduce."""
        db, store = get_connection()
        args = [inpd[k] for k in arg_names]
        split_data = [constant_artefact(db, store, d) for d in split_fun(*args)]
        apply_data = [apply_fun(d) for d in split_data]
        combine_data = combine_fun(apply_data)
        return dict(out=combine_data)

    # Generate the subdag operations
    cmd = subdag_funsie(__sac, inp_types, {"out": out}, name=fun_name, strict=strict)
    operation = make_op(db, cmd, inputs, opt)
    return Artefact.grab(db, operation.out["out"])



"""Install funsies."""
import setuptools
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="funsies",
    version="0.8.1",
    author="Cyrille Lavigne",
    author_email="cyrille.lavigne@mail.utoronto.ca",
    description="Funsies is a library to build and execution engine for"
    + " reproducible, composable and data-persistent computational workflows.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aspuru-guzik-group/funsies",
    package_dir={"": "src"},
    package_data={"funsies": ["py.typed"]},  # mypy exports
    packages=setuptools.find_namespace_packages(where="src"),
    # Dependencies
    python_requires=">=3.7",
    install_requires=[
        "mypy_extensions",
        "redis",
        "cloudpickle",
        "rq>=1.7",
        "loguru",
        'importlib-metadata ~= 1.0 ; python_version < "3.8"',
        'typing_extensions ; python_version < "3.8"',
        "chevron",
    ],
    entry_points="""
        [console_scripts]
        funsies=funsies._cli:main
        start-funsies=funsies._start_funsies:main
    """,
    classifiers=[
        "Development Status :: 4 - Beta",
        #
        "Typing :: Typed",
        #
        "License :: OSI Approved :: MIT License",
        #
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        #
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    keywords="workflows hashtree redis compchem chemistry parallel hpc",
)



"""Run DAGs from DAGs using funsies."""
from __future__ import annotations

# std
from io import BytesIO
import time
from typing import Any, Callable, Mapping, Optional

# external
import cloudpickle

# module
from ._constants import Encoding
from ._funsies import Funsie, FunsieHow
from ._graph import Artefact
from ._logging import logger
from .errors import Result

# types
subdag_t = Callable[..., Any]


def subdag_funsie(
    fun: subdag_t,
    inputs: dict[str, Encoding],
    outputs: dict[str, Encoding],
    name: Optional[str] = None,
    strict: bool = True,
) -> Funsie:
    """Wrap a python function that creates a DAG.

    The callable should have the following signature:

        f(dict[str, Artefact]) -> dict[str, Artefact]

    This is similar to python_funsie in _pyfunc.py

    Args:
        fun: a Python callable f(inp)->out.
        inputs: Keys in the input dictionary inp.
        outputs: Keys in the output dictionary out.
        name:
            Name of callable. If not given, the qualified name of the callable
            is used instead.
        strict:
            If True (default), the function will not run if any of its inputs
            are Error-ed.

    Returns:
        A Funsie instance.
    """
    if name is None:
        name = callable.__qualname__

    ierr = 1
    if strict:
        ierr = 0

    return Funsie(
        how=FunsieHow.subdag,
        what=name,
        inp=inputs,
        out=outputs,
        error_tolerant=ierr,
        extra={"pickled function": cloudpickle.dumps(fun)},
    )


def run_subdag_funsie(
    funsie: Funsie, input_values: Mapping[str, Result[BytesIO]]
) -> dict[str, Optional[Artefact[Any]]]:
    """Execute a subdag generator."""
    logger.info("subdag generator")
    fun: subdag_t = cloudpickle.loads(funsie.extra["pickled function"])
    name = funsie.what
    logger.info(f"$> {name} subdag generator")

    decoded = funsie.decode(input_values)

    # run
    t1 = time.time()
    outfun = fun(decoded)
    t2 = time.time()

    logger.info(f"done 1/1 \t\tduration: {t2-t1:.2f}s")
    out: dict[str, Optional[Artefact[Any]]] = {}
    for output in funsie.out:
        if output in outfun:
            out[output] = outfun[output]
        else:
            logger.warning(f"missing expected output {output}")
            out[output] = None

    return out



"""Some useful functions for workflows."""
from __future__ import annotations

# std
from typing import Any, Callable, Optional, Sequence, TypeVar, Union

# external
from redis import Redis

# module
from ._constants import _Data, Encoding
from ._graph import Artefact
from ._shell import ShellOutput
from .config import Options
from .errors import Error, Result
from .fp import morph, py, reduce
from .ui import execute, wait_for

_TargetBytes = Union[Artefact[bytes], bytes]

Tin = TypeVar("Tin")
Tout1 = TypeVar("Tout1")
Tout2 = TypeVar("Tout2")
Td = TypeVar("Td", bound=_Data)


def execute_all(
    things: Sequence[Union[Artefact[Any], ShellOutput]],
    block: bool = True,
    timeout: Optional[float] = None,
    *,
    opt: Optional[Options] = None,
    connection: Optional[Redis[bytes]] = None,
) -> Artefact[bytes]:
    """Execute a sequence of artefacts or shell commands.

    `execute_all()` executes a sequence of artefact in arbitrary, parallel
    order. It's equivalent to the following simple program,

    ```python
    for el in things:
        funsies.execute(el)

    for el in things:
        funsies.wait_for(el)
    ```

    The return value is `funsies.errors.Result` containing an empty bytestring
    or `funsies.errors.Error` if any of `things` is `funsies.errors.Error`.

    Args:
        things: A bunch of Artefact or ShellOutput.
        block: Whether to block until all artefacts have been executed. Defaults to True.
        timeout: `timeout=` argument for `funsies.wait_for()` when blocking.
        connection: An explicit Redis connection. Not required if called
            within a `funsies.Fun()` context.
        opt: An `funsies.config.Options` instance generated from
            `funsies.options()`. Not required if called within a
            `funsies.Fun()` context.

    Returns:
        An `funsies.types.Artefact[bytes]` object that resolves to `b""`.
    """
    things2 = [t if isinstance(t, Artefact) else t.stdout for t in things]
    out = reduce(
        lambda *x: b"", *things2, out=Encoding.blob, opt=opt, connection=connection
    )
    execute(out, connection=connection)
    if block:
        wait_for(out, timeout=timeout, connection=connection)
    return out


def match_results(
    results: Sequence[Result[Tin]],
    some: Callable[[Tin], Tout1],
    error: Optional[Callable[[Error], Tout2]] = None,
) -> list[Union[Tout1, Tout2]]:
    """Match on result errors."""
    out: list[Union[Tout1, Tout2]] = []
    for el in results:
        if isinstance(el, Error):
            if error is not None:
                out += [error(el)]
        else:
            out += [some(el)]
    return out


def concat(
    *inp: _TargetBytes,
    join: Union[str, bytes] = b"",
    strip: bool = False,
    strict: bool = True,
    opt: Optional[Options] = None,
    connection: Optional[Redis[bytes]] = None,
) -> Artefact:
    """Concatenate artefacts."""

    def concatenation(joiner: bytes, do_strip: bool, *args: Result[bytes]) -> bytes:
        lines = match_results(args, lambda x: x)
        out = b""
        for i, l in enumerate(lines):
            if do_strip:
                out += l.strip()
            else:
                out += l

            if i != len(lines) - 1:
                out += joiner
        return out

    # type convert str to bytes
    if isinstance(join, str):
        join = join.encode()

    return py(  # type:ignore
        concatenation,
        join,
        strip,
        *inp,
        out=[Encoding.blob],
        strict=strict,
        connection=connection,
        opt=opt,
    )


def truncate(
    inp: _TargetBytes,
    top: int = 0,
    bottom: int = 0,
    separator: Union[str, bytes] = b"\n",
    opt: Optional[Options] = None,
    connection: Optional[Redis[bytes]] = None,
) -> Artefact:
    """Truncate an artefact."""

    def __truncate(inp: bytes, top: int, bottom: int, sep: bytes) -> bytes:
        data = inp.split(sep)
        i = top
        j = len(data) - bottom
        return sep.join(data[i:j])

    # type convert str to bytes
    if isinstance(separator, str):
        separator = separator.encode()

    return reduce(
        __truncate,
        inp,
        top,
        bottom,
        separator,
        out=Encoding.blob,
        name="truncate",
        strict=True,
        opt=opt,
        connection=connection,
    )


def stop_if(
    fun: Callable[[Td], bool],
    inp: Union[Td, Artefact[Td]],
    opt: Optional[Options] = None,
    connection: Optional[Redis[bytes]] = None,
) -> Artefact[Td]:
    """Stop execution if a condition holds."""

    def __stop_if(inp: Td) -> Td:
        if fun(inp):
            raise RuntimeError("Data triggered stop.")
        else:
            return inp

    fun_name = f"stop_if:{fun.__qualname__}"
    return morph(
        __stop_if, inp, name=fun_name, strict=True, connection=connection, opt=opt
    )


def not_empty(
    inp: _TargetBytes,
    opt: Optional[Options] = None,
    connection: Optional[Redis[bytes]] = None,
) -> Artefact:
    """Stop DAG on empty files."""

    def __not_empty(inp: bytes) -> bytes:
        if len(inp):
            return inp
        else:
            raise RuntimeError("This file is empty.")

    fun_name = "not an empty file"
    return py(
        __not_empty, inp, name=fun_name, strict=True, connection=connection, opt=opt
    )



"""Engines for storing and retrieving artefacts."""
from __future__ import annotations

# std
from io import BytesIO
import os
import traceback
from typing import cast, NewType, Optional

# external
from redis import Redis

# module
from ._constants import _AnyPath, ARTEFACTS, hash_t, join
from ._logging import logger
from .errors import Error, ErrorKind, Result

descr_t = NewType("descr_t", str)


class StorageEngine:
    """Baseclass implementing the artefact storage protocol."""

    def get_key(self: StorageEngine, hash: hash_t) -> descr_t:
        """Return the key for a given hash."""
        raise NotImplementedError("Baseclass used where derived class is required.")

    def take(self: StorageEngine, key: descr_t) -> Result[BytesIO]:
        """Return a bytes stream for a given key.

        Note: It is the caller's responsibility to .close() the stream.
        """
        raise NotImplementedError("Baseclass used where derived class is required.")

    def put(self: StorageEngine, key: descr_t, data: BytesIO) -> Optional[Error]:
        """Write stream data to a given key.

        Note: this does not .close() the stream.
        """
        raise NotImplementedError("Baseclass used where derived class is required.")


# --------------------------------------------------------------------------
# Storage engine configurations
# --------------------------------------------------------------------------

# Data storage on disk
class DiskStorage(StorageEngine):
    """Storage engine that puts artefact data on disk."""

    def __init__(self: DiskStorage, path: _AnyPath) -> None:
        """Return a storage engine that saves to a given path."""
        self.path = os.path.abspath(path)
        logger.info(f"saving artefacts to {self.path}")
        os.makedirs(self.path, exist_ok=True)
        self.buffer_length = 16 * 1024

    def get_key(self: DiskStorage, hash: hash_t) -> descr_t:
        """Return the key for a given hash."""
        dir = os.path.join(self.path, hash[:2])
        os.makedirs(dir, exist_ok=True)
        key = os.path.join(dir, hash[2:])
        return descr_t(str(key))

    def take(self: DiskStorage, key: descr_t) -> Result[BytesIO]:
        """Return a bytes stream for a given key.

        Note: It is the caller's responsibility to .close() the stream.
        """
        try:
            fdst = cast(BytesIO, open(key, "rb"))
        except Exception:
            tb_exc = traceback.format_exc()
            return Error(
                kind=ErrorKind.ExceptionRaised,
                details=tb_exc,
            )

        return fdst

    def put(self: DiskStorage, key: descr_t, data: BytesIO) -> Optional[Error]:
        """Write stream data to a given key.

        Note: this function does not .close() the stream.
        """
        try:
            with open(key, "wb") as fdst:
                while True:
                    buf = data.read(self.buffer_length)
                    if not buf:
                        break
                    fdst.write(buf)

        except Exception:
            tb_exc = traceback.format_exc()
            return Error(
                kind=ErrorKind.ExceptionRaised,
                details=tb_exc,
            )
        return None


_DEFAULT_BLOCK_SIZE = 30 * 1024 * 1024  # 30 MB


class RedisStorage(StorageEngine):
    """Storage engine that puts artefact data in Redis."""

    def __init__(
        self: RedisStorage,
        conn: Redis[bytes],
        block_size: int = _DEFAULT_BLOCK_SIZE,
    ) -> None:
        """Returns a storage engine that saves to a Redis instance."""
        self.instance = conn
        self.block_size = block_size

    def get_key(self: RedisStorage, hash: hash_t) -> descr_t:
        """Return the key for a given hash."""
        key = join(ARTEFACTS, hash, "data")
        return descr_t(key)

    def take(self: RedisStorage, key: descr_t) -> Result[BytesIO]:
        """Return a bytes stream for a given key.

        Note: It is the caller's responsibility to .close() the stream.
        """
        # TODO: stream it?
        # TODO: check for truncation
        out = self.instance.lrange(key, 0, -1)
        if len(out) == 0:
            return Error(
                ErrorKind.DataNotFound,
                details=f"No data at address {key} in redis instance.",
            )
        else:
            return BytesIO(b"".join(out))

    def put(self: RedisStorage, key: descr_t, data: BytesIO) -> Optional[Error]:
        """Write stream data to a given key.

        Note: this function does not .close() the stream.
        """
        first = True  # this workaround is to make sure that writing no data is ok.
        pipe = self.instance.pipeline(transaction=True)
        pipe.delete(key)
        while True:
            try:
                dat = data.read(self.block_size)
            except Exception:
                tb_exc = traceback.format_exc()
                return Error(
                    kind=ErrorKind.ExceptionRaised,
                    details=tb_exc,
                )

            if len(dat) == 0 and not first:
                break
            else:
                pipe.rpush(key, dat)

            first = False
        pipe.execute()
        return None



"""Contextual DB usage."""
from __future__ import annotations

# std
from contextlib import contextmanager
from dataclasses import replace
import shutil
import subprocess
import tempfile
import time
from typing import Any, Iterator, Optional, Sequence, Tuple, Union

# external
from redis import Redis
import rq
from rq import command, Worker
from rq.local import LocalStack

# module
from ._constants import _AnyPath, hash_t, join, OPERATIONS
from ._logging import logger
from ._storage import StorageEngine
from .config import Options, Server

# A thread local stack of connections (adapted from RQ)
_options_stack = LocalStack()
_storage_stack = LocalStack()


def cleanup_funsies(db: Redis[bytes]) -> None:
    """Clean up Redis instance of DAGs, Queues and clear workers."""
    queues = rq.Queue.all(connection=db)
    for queue in queues:
        queue.delete(delete_jobs=True)

    # Reset operation status
    ops = join(OPERATIONS, hash_t("*"), "owner")
    keys = db.keys(ops)
    if len(keys):
        logger.info(f"clearing {len(keys)} unfinished ops")
        for k in keys:
            db.delete(k)


def shutdown_workers(db: Redis[bytes], force: bool) -> None:
    """Shutdown funsies workers."""
    workers = Worker.all(db)
    logger.info(f"shutting down {len(workers)} workers")
    for worker in workers:
        command.send_shutdown_command(db, worker.name)  # Tells worker to shutdown
        if force:
            command.send_kill_horse_command(db, worker.name)


# --------------------------------------------------------------------------------


def get_redis(db: Optional[Redis[bytes]] = None) -> Redis[bytes]:
    """Get Redis instance."""
    if db is None:
        myjob = rq.get_current_job()
        connect: Optional[Redis[bytes]] = rq.connections.get_current_connection()
        if connect is not None:
            return connect
        elif myjob is not None:
            out2: Redis[bytes] = myjob.connection
            return out2
        else:
            raise RuntimeError("No redis instance available.")
    else:
        return db


def get_storage(store: Optional[StorageEngine] = None) -> StorageEngine:
    """Get current storage method."""
    if store is not None:
        return store
    else:
        if _storage_stack.top is not None:
            out: StorageEngine = _storage_stack.top
            return out
        else:
            raise RuntimeError("No Storage instance available.")


def get_options(opt: Optional[Options] = None) -> Options:
    """Get Options instance."""
    if opt is not None:
        return opt
    else:
        if _options_stack.top is not None:
            out: Options = _options_stack.top
            return out
        else:
            raise RuntimeError("No Options instance available.")


def options(**kwargs: Any) -> Options:
    """Set operation and workflow options.

    This function sets specific configuration options for an operation or a
    workflow that do not change hash values or cause re-execution, but do
    change runtime behaviour, such as job timeouts, queue selection, etc.
    Available options and their names are described in the entry for
    `config.Options`.

    This function wraps the `config.Options` with layering of default values.
    The value of each attribute of `config.Options` is set based on:

    1. The values set by the `**kwargs` dictionary.

    2. The values set in the `**kwargs` dictionary of the enclosing
    `funsies.Fun()` context, if `default=options(**kwargs)` is passed to
    `funsies.Fun()`.

    3. The default values in `config.Options`.

    This allows layering of fairly complex runtime behaviour.

    """
    if _options_stack.top is None:
        return Options(**kwargs)
    else:
        defaults: Options = _options_stack.top
        return replace(defaults, **kwargs)


Connection = Union[
    Redis,
    StorageEngine,
    Tuple[Redis, StorageEngine],
    Tuple[StorageEngine, Redis],
    None,
]
"""A convenience type for information about a funsies server setup."""


def get_connection(inp: Connection = None) -> tuple[Redis[bytes], StorageEngine]:
    """Destructure a Connection."""
    db: Optional[Redis[bytes]] = None
    st = None
    if isinstance(inp, tuple):
        if isinstance(inp[0], Redis):
            db = inp[0]  # type:ignore
            assert isinstance(inp[1], StorageEngine)
            st = inp[1]
        elif isinstance(inp[0], StorageEngine):
            assert isinstance(inp[1], Redis)
            db = inp[1]  # type:ignore
            st = inp[0]
        else:
            raise RuntimeError(f"Wrong input for get_connection, {inp}")
    else:
        if isinstance(inp, Redis):
            db = inp  # type:ignore
        elif isinstance(inp, StorageEngine):
            st = inp
        elif inp is None:
            pass
        else:
            raise RuntimeError(f"Wrong input for get_connection, {inp}")

    return get_redis(db), get_storage(st)


# ---------------------------------------------------------------------------------
# Contexts
@contextmanager
def Fun(
    server: Optional[Server] = None,
    defaults: Optional[Options] = None,
    cleanup: bool = False,
) -> Iterator[Redis[bytes]]:
    """Context manager for redis connections."""
    if server is None:
        logger.warning("Opening new redis connection with default settings...")
        server = Server()

    if defaults is None:
        defaults = Options()
    _options_stack.push(defaults)

    db, store = server.new_connection()
    _storage_stack.push(store)

    if cleanup:
        cleanup_funsies(db)

    # also push on rq
    rq.connections.push_connection(db)

    try:
        yield rq.connections.get_current_connection()
    finally:
        popped = rq.connections.pop_connection()
        assert popped == db, (
            "Unexpected Redis connection was popped off the stack. "
            "Check your Redis connection setup."
        )
        _ = _options_stack.pop()


@contextmanager
def ManagedFun(
    nworkers: int = 1,
    worker_args: Optional[Sequence[str]] = None,
    redis_args: Optional[Sequence[str]] = None,
    defaults: Optional[Options] = None,
    directory: Optional[_AnyPath] = None,
    data_url: Optional[str] = None,
    pw: Optional[str] = None,
) -> Iterator[Redis[bytes]]:
    """Make a fully managed funsies db."""
    if directory is None:
        dir = tempfile.mkdtemp()
    else:
        dir = str(directory)

    logger.debug(f"running redis-server in {dir}")

    if worker_args is not None:
        wargs = [el for el in worker_args]
    else:
        wargs = []

    if redis_args is not None:
        rargs = [el for el in redis_args]
    else:
        rargs = []

    if pw is None:
        # std
        import secrets

        pw = secrets.token_hex(12)

    # Start redis
    port = 16379
    url = f"redis://:{pw}@localhost:{port}"
    cmdline = ["redis-server"] + rargs + ["--port", f"{port}", "--requirepass" f" {pw}"]
    if data_url is None:
        data_url = url
    server = Server(jobs_url=url, data_url=data_url)

    redis_server = subprocess.Popen(
        cmdline,
        cwd=dir,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    time.sleep(0.1)
    stat = redis_server.poll()
    if stat is not None:
        assert redis_server.stderr is not None
        stdout = redis_server.stderr.read().decode()
        raise RuntimeError(f"Redis server failed to start, errcode={stat}\n{stdout}")

    logger.debug(f"redis running at {url}")

    # spawn workers
    logger.debug(f"spawning {nworkers} funsies workers")
    worker_pool = [
        subprocess.Popen(
            ["funsies", "--jobs", url, "--data", data_url, "worker"] + wargs, cwd=dir
        )
        for i in range(nworkers)
    ]

    try:
        logger.success(f"{nworkers} workers connected to {url.split('@')[-1]}")
        with Fun(server=server, defaults=defaults) as db:
            db.set("bla", 3)
            yield db
    finally:
        logger.debug("terminating worker pool and server")
        for w in worker_pool:
            w.kill()
            w.wait()
        # stop db
        db.shutdown()  # type:ignore
        db.connection_pool.disconnect()
        redis_server.wait()
        if directory is None:
            shutil.rmtree(dir)
        logger.success("stopping managed fun")
